"""Test CSV"""

from luminadb import Database
from luminadb.csv import to_csv_file, to_csv_string

from ..setup import setup_database, pstdout, temp_dir

def test_export_csv():
    """Test 0600 Export to CSV"""
    database = Database(":memory:")
    setup_database(database)
    csv = to_csv_string(database)
    print("test_07_00_export_csv\n", csv, "\n", file=pstdout, flush=True)
    assert csv


def test_export_file():
    """Test 0701 Export to CSV file"""
    database = Database(":memory:")
    setup_database(database)
    assert to_csv_file(database.table("users"), temp_dir / "users.csv")  # type: ignore


def test_export_directory():
    """Test 0702 Export to CSV as directory"""
    database = Database(":memory:")
    setup_database(database)
    assert to_csv_file(database, temp_dir / "MemDB")  # type: ignore



"""Table API delete tests"""

from luminadb import Database
from ..setup import setup_database_builder, setup_database_fns, USER_BASE

def test_delete():
    """Test 0600 Deletion test"""
    db = Database(":memory:")
    setup_database_builder(db)
    users = db.table("users")
    assert users.delete() == len(USER_BASE)


def test_delete_condition():
    """Test 0601 Delete based on ID"""
    db = Database(":memory:")
    setup_database_builder(db)
    users = db.table("users")
    assert users.delete({"id": 1}) == 1


def test_delete_limit():
    """Test 0602 Delete with limit"""
    db = Database(":memory:")
    setup_database_builder(db)
    users = db.table("users")
    assert users.delete(limit=1) == 1


def test_delete_limit_condition():
    """Test 0603 Delete limit with certain condition"""
    db = Database(":memory:")
    setup_database_fns(db)
    checkout = db.table("checkout")
    assert checkout.delete({"quantity": 50}, limit=2) == 2


def test_delete_order():
    """Test 0603 Delete limit with certain condition"""
    db = Database(":memory:")
    setup_database_fns(db)
    checkout = db.table("checkout")
    assert checkout.delete({"quantity": 50}, order="asc", limit=2) == 2  # type: ignore



"""User Benchmark

This module is intended as a standing point for anyone who wants to contribute to the project.
Although, any kind of help is appreciated ;)

This module however, is not shipped to the production and only be fetched manually by cloning
the project.

To get started, this module is intended to be imported. So opens up your python interpreter.

-- Details

This module has a database connection from "testdb.sqlite3" which assigned as database
Then, a table person has columns of text name, integer age, and text gender.

The module also have function fills to fill the table with contents to whatever amount you want.
"""
# pylint: disable=wrong-import-position
# Edit this if you get a error.
from random import choice
from sys import path
from os.path import realpath
path.insert(0, realpath("../"))

from luminadb import Database, text, integer
from luminadb._debug import STATE


STATE['DEBUG'] = True
database = Database("testdb.sqlite3")

person = database.create_table("person", [text("name"), integer("age"), text("gender")])


def fills(amount):
    """Directly filled the table person with pre-generated contents."""
    data = []
    names = ("John Doe", "Felina")
    age = range(100)
    for i in range(amount):
        _pname = choice(names)
        person_name = f"{_pname} #{i}"
        person_age = choice(age)
        person_gender = "M" if _pname == names[0] else "F"
        data.append({"name": person_name, "age": person_age, "gender": person_gender})
    person.insert_many(data)
    return amount



"""Model helpers"""

# pylint: disable=invalid-name,too-few-public-methods,abstract-method,protected-access

from typing import Any, Callable, Type, TypeAlias, TypeVar, overload
from enum import StrEnum

import luminadb
from .errors import ValidationError
from ..column import BuilderColumn, text, integer, blob, boolean, real

TypeFunction: TypeAlias = Callable[[str], BuilderColumn]
Model = TypeVar("Model", bound="luminadb.BaseModel")
FuncT = TypeVar("FuncT", bound=Callable[..., bool])
BaseModel: TypeAlias = "luminadb.BaseModel"

TYPES: dict[Type[Any], TypeFunction] = (
    {  # pylint: disable=possibly-used-before-assignment
        int: integer,
        str: text,
        bytes: blob,
        bool: boolean,
        float: real
    }
)

VALID_HOOKS_NAME = (
    "before_create",
    "after_create",
    "before_update",
    "after_update",
    "before_delete",
    "after_delete",
)


class ConstraintEnum(StrEnum):
    """Constraints for update/delete"""

    RESTRICT = "restrict"
    SETNULL = "null"
    CASCADE = "cascade"
    NOACT = "no act"
    DEFAULT = "default"


RESTRICT = ConstraintEnum.RESTRICT
SETNULL = ConstraintEnum.SETNULL
CASCADE = ConstraintEnum.CASCADE
NOACT = ConstraintEnum.NOACT
DEFAULT = ConstraintEnum.DEFAULT


class Constraint:
    """Base constraint class for models"""

    def __init__(self, column: str) -> None:
        self._column = column

    @property
    def column(self):
        """Columns"""
        return self._column

    def apply(self, type_: BuilderColumn):
        """Apply this constraint to an column"""
        raise NotImplementedError()


class Unique(Constraint):
    """Unique constraint"""

    def apply(self, type_: BuilderColumn):
        type_.unique()


class Foreign(Constraint):
    """Foreign constraint"""

    def __init__(self, column: str, target: str | Type[Model]) -> None:
        super().__init__(column)
        self._target = target
        self.resolve()
        self._base = target
        self._on_delete = DEFAULT
        self._on_update = DEFAULT

    @property
    def target(self):
        """Target foreign constraint"""
        return self._target

    def on_delete(self, constraint: ConstraintEnum):
        """On delete constraint"""
        self._on_delete = constraint
        return self

    def on_update(self, constraint: ConstraintEnum):
        """On update constraint"""
        self._on_update = constraint
        return self

    def resolve(self):
        """Resolve if current target is a Model"""
        if issubclass(self._target, luminadb.BaseModel):  # type: ignore
            name = self._target.__table_name__
            target = self._target._primary  # pylint: disable=protected-access
            if not target:
                raise ValueError(f"{type(self._target)} does not have primary key")
            self._target = f"{name}/{target}"

    def apply(self, type_: BuilderColumn):
        type_.foreign(self._target)  # type: ignore
        if self._on_delete != DEFAULT:
            type_.on_delete(self._on_delete.value)
        if self._on_update != DEFAULT:
            type_.on_update(self._on_update.value)


class Primary(Constraint):
    """Primary constraint"""

    def apply(self, type_: BuilderColumn):
        """Apply this constraint as primary"""
        type_.primary()


class Validators:
    """Base class to hold validators"""

    def __init__(self, fn: Callable[[Any], bool], if_fail: str) -> None:
        self._callable = fn
        self._reason = if_fail

    def validate(self, instance: BaseModel):
        """Validate a value"""
        if not self._callable(instance):
            err = ValidationError(self._reason)
            err.add_note(str(instance))
            raise err
        return True


@overload
def hook(fn_or_name: Callable[[Model], None]) -> "staticmethod[[Callable[[Model], None]], None]":
    pass


@overload
def hook(fn_or_name: str):
    pass


def hook(fn_or_name):
    """Register a hook"""

    def decorator(func):
        fn = staticmethod(func)
        fn_name = func.__name__
        final_name = fn_or_name if fn_name not in VALID_HOOKS_NAME else fn_name

        if final_name is None:
            raise ValueError("Hooks name is not valid. Provide with @hook(name)")
        fn._hooks_info = (fn_or_name  # type: ignore
                          if fn_name not in VALID_HOOKS_NAME else fn_name,)
        return fn

    return decorator(fn_or_name) if callable(fn_or_name) else decorator


@overload
def validate(fn_or_column: FuncT) -> "staticmethod[[FuncT], bool]":
    pass


@overload
def validate(fn_or_column: str, reason: str | None = None):  # type: ignore
    pass


def validate(fn_or_column, reason=None):
    """Register a validator"""

    def decorator(func: Callable):
        print(func, fn_or_column)
        fn = staticmethod(func)
        name = func.__name__
        inferred_col = (
            name[len("validate_") :] if name.startswith("validate_") else None
        )

        col = fn_or_column or inferred_col
        if callable(col):
            col = inferred_col
        if col is None:
            raise ValueError("Validator must have a column name.")

        msg = reason or func.__doc__ or f"Validation failed for '{col}'"
        fn._validators_info = (col, msg)  # type: ignore
        return fn

    return decorator(fn_or_column) if callable(fn_or_column) else decorator


def initiate_hook(cls: Type[BaseModel], member: Callable):
    """Initiate hooks"""
    if hasattr(member, "_hooks_info"):
        info = member._hooks_info
        cls._register("hook", info[0])(member)


def initiate_validators(cls: Type[BaseModel], member: Callable):
    """Initiate validators"""
    if hasattr(member, "_validators_info"):
        info = member._validators_info
        cls._register("validator", info[0], info[1])(member)



"""Test Custom"""

from luminadb import Database, this
from luminadb.operators import in_
from luminadb._debug import STATE

from .setup import setup_orderable, file, pstdout

def test_98_00_test():
    """Gradual test"""
    db = Database(":memory:")
    setup_orderable(db)
    items = db.table("items")
    v0 = items.select({"quantity": this == 99})
    STATE["DEBUG"] = True
    values = items.select({"quantity": in_([99, 98, 97])})
    assert v0
    assert values
    STATE["DEBUG"] = False


def test_99_99_save_report():
    """FINAL 9999 Save reports"""
    with open(file, "w", encoding="utf-8") as xfile:
        xfile.write(pstdout.getvalue())



"""Test other features"""

from sqlite3 import OperationalError
from random import randint

from pytest import raises
from luminadb import Database, integer, text
# from luminadb.workers import DatabaseWorker
from luminadb.index import Index

from ..setup import setup_database_fns, setup_database, count


def test_function_count():
    """Test 0800 Count(*) usage"""
    database = Database(":memory:")
    setup_database_fns(database)
    counted = count("*")
    data = database.table("checkout").select(what=counted)
    print(data)
    assert data == 4


def test_select_error():
    """Test errors"""
    db = Database(":memory:")
    setup_database(db)
    groups = db.table("groups")
    with raises(OperationalError):
        groups.select({"nothing": None})


def test_pragma_foreign_key():
    """Test 1001 PRAGMA tests"""
    db = Database(":memory:")
    setup_database(db)
    db.foreign_pragma("ON")
    db.foreign_pragma("OFF")
    assert db.foreign_pragma() == {"foreign_keys": 0}


def test_pragma_optimize():
    """Test 1002 PRAGMA optimize"""
    db = Database(":memory:")
    setup_database(db)
    db.optimize()
    assert 1 == 1


def test_pragma_shrink():
    """Test 1003 PRAGMA shrink"""
    db = Database(":memory:")
    t = db.create_table("t", [integer("a")])
    t.insert_many([{"a": a} for a in range(10000)])
    t.commit()
    db.shrink_memory()
    assert 1 == 1


def test_vacuum():
    """Test 1004 vacuum"""
    db = Database(":memory:")
    t = db.create_table("t", [integer("a")])
    t.insert_many([{"a": a} for a in range(10000)])
    t.commit()
    _ = [t.delete_one({"a": randint(0, 1000)})]
    t.commit()
    db.vacuum()


# @mark.skipif(sys.version_info < (3, 13), reason="Worker feature is exclusive in 3.13")
# def test_worker():
#     """Test worker"""

#     db = DatabaseWorker(":memory:")
#     t = db.create_table("t", [integer("a")])
#     t.insert({"a": 1})
#     t.commit()
#     assert t.select_one({"a": 1}) is not None
#     db.close()


def test_index():
    """Test index"""

    db = Database(":memory:")
    t = db.create_table("t", [integer("a"), text("b")])
    index = Index('testindex').target('t').columns('a', 'b')
    assert db.create_index(index) is None, "Index should be created"
    t.select_one()
    assert db.delete_index(index) is None, "Index should be destroyed"



"""Model QueryBuilder"""

# pylint: disable=protected-access

from __future__ import annotations
from typing import Any, Generic, Type, TypeVar
from .errors import NoDataReturnedError
from ..column import check_one
from ..functions import Function
from .. import models  # pylint: disable=unused-import

T = TypeVar("T", bound="models.BaseModel")
count = Function("COUNT")

MGR_REQ = "Note: The database record has serious mismatch, consider doing migration."

def add_excnote(exc: Exception):
    """Add exception notes when an entry retrieved is different"""
    msg = str(exc)
    if "() missing" in msg or "() got" in msg:
        exc.add_note(MGR_REQ)
    # if "() got" in msg:
        # exc.add_note(MGR_REQ)
class QueryBuilder(Generic[T]):
    # pylint: disable=protected-access
    """Query builder for Model ORM"""

    def __init__(self, model: Type[T]) -> None:
        self._model = model
        self._filters: dict[str, Any] = {}
        self._limit = 0
        self._offset = 0
        self._order = None
        self._failing = False

    def throw(self):
        """Set when fetch() returns nothing, will raise an error"""
        self._failing = True
        return self

    def where(self, **kwargs: Any):
        """Sets conditioning"""
        self._filters.update(kwargs)
        return self

    def limit(self, value: int):
        """Sets limit"""
        self._limit = value
        return self

    def offset(self, value: int):
        """Sets offset"""
        self._offset = value
        return self

    def order_by(self, column: str, descending: bool = False):
        """Order the query by a column"""
        self._order = (check_one(column), "asc" if descending is False else "desc")
        return self

    def fetch(self) -> list[T]:
        """Fetch data from table"""
        records = self._model._tbl.select(  # pylint: disable=protected-access
            self._filters, limit=self._limit, offset=self._offset, order=self._order
        )

        if len(records) == 0 and self._failing:
            raise NoDataReturnedError(
                f"Model {self._model.__name__} has no data for current scope"
            )
        try:
            return [self._model(**record) for record in records]
        except TypeError as exc:
            add_excnote(exc)
            raise exc

    def fetch_one(self) -> T | None:
        """Fetch one data from table"""
        # pylint: disable=protected-access
        record = self._model._tbl.select_one(self._filters, order=self._order)
        if not record and self._failing:
            raise NoDataReturnedError(
                f"Model {self._model.__name__} has no data for current scope"
            )
        if not record:
            return None
        try:
            return self._model(**record)
        except TypeError as exc:
            add_excnote(exc)
            raise exc

    def patch(self, **kwargs: Any):
        """Update a data based on the filter according to passed keyword args"""
        affected_rows = self._model._tbl.update(
            self._filters, kwargs, self._limit, self._order
        )

        if affected_rows == 0 and self._failing:
            raise NoDataReturnedError(
                f"Model {self._model.__name__} updates no data for current scope"
            )
        return affected_rows

    def delete(self):
        """Delete data based on the filter"""
        affected_rows = self._model._tbl.delete(self._filters, self._limit, self._order)

        if affected_rows == 0 and self._failing:
            raise NoDataReturnedError(
                f"Model {self._model.__name__} delete exactly no rows"
            )

        return affected_rows

    def count(self) -> int:
        """Count how much data is within this operation"""
        return self._model._tbl.select(self._filters, what=count("*"))

    update = patch



"""Locals"""

from re import compile as re_compile
from re import escape as re_escape
from string import punctuation
from typing import Literal

_NO_QUOTED_STR = "'\";"
_NO_QOUTED = re_compile(f"[{re_escape(_NO_QUOTED_STR)}]+")
_NO_UNLIKE_STR = punctuation.replace("%", "").replace("_", "")
_NO_UNLIKE = re_compile(f"[{re_escape(_NO_UNLIKE_STR)}]+")
_NPATH_STR = punctuation.replace("/", "")
_NPATH = re_compile(f"[{re_escape(_NPATH_STR)}]+")
_PATH = re_compile(f"[{re_escape('/')}]+")

PLUGINS_PATH = ("--mysql",)

_SQLITETYPES = ["blob", "null", "integer", "real", "text"]

SQLACTION = (
    Literal["null"]
    | Literal["cascade"]
    | Literal["no act"]
    | Literal["default"]
    | Literal["restrict"]
)



"""Error modules"""

from ..errors import UnexpectedResultError

class ValidationError(ValueError):
    """Specific validation error; a value fails validation test"""

class ConstraintError(UnexpectedResultError):
    """Specific constraint is found twice, thrice, etc."""

class NoDataReturnedError(ValueError):
    """Data fetched from model cannot be retrieved"""



"""Generic query building engine"""

from typing import Optional, Any
from dataclasses import dataclass
from functools import lru_cache

from .typings import CacheCond, OnlyColumn, CacheOrders, CacheData, Condition, SubQuery
from .utils import (
    parse_orders,
    format_paramable,
    setup_limit_patch,
    MAX_SUBQUERY_STACK_LIMIT,
    NAMING_FORMAT,
)

from ..errors import SecurityError
from ..functions import ParsedFn
from ..signature import Signature
from ..utils import check_one, check_iter, null
from ..utils import generate_ids


@dataclass(frozen=True)
class QueryParams:
    """Encapsulates parameters for building SQL queries."""

    table_name: str
    condition: Optional[CacheCond] = None
    only: Optional[OnlyColumn] = None
    limit: int = 0
    offset: int = 0
    order: Optional[CacheOrders] = None
    data: Optional[CacheData] = None

    def __post_init__(self):
        if not all(
            (isinstance(getattr(self, item), int) for item in ("limit", "offset"))
        ):
            raise TypeError("Expected limit/offset to be integer")

    def __hash__(self):
        """Custom hash function to ensure compatibility with lru_cache."""
        return hash(
            (
                self.table_name,
                self.condition,
                self.only,
                self.limit,
                self.offset,
                self.order,
                self.data,
            )
        )


@lru_cache
def _build_select(query_params: QueryParams, depth: int = 0):
    if depth < 0 or depth >= MAX_SUBQUERY_STACK_LIMIT:
        raise RecursionError(
            "Subquery builder has reached recursion limit of"
            f"{MAX_SUBQUERY_STACK_LIMIT}"
        )
    check_one(query_params.table_name)
    cond, data = extract_signature(query_params.condition, depth=depth)
    what_ = "*"
    if query_params.only and isinstance(query_params.only, ParsedFn):
        what_, databin = query_params.only.parse_sql()
        check_iter(
            (query_params.only.name, *(a for a in query_params.only.values if a != "*"))
        )
        data.update(databin)
    elif isinstance(query_params.only, tuple):
        generator = (
            column_name for column_name in query_params.only if check_one(column_name)
        )
        what_ = f"{', '.join(generator)}"  # type: ignore
    elif query_params.only != "*" and isinstance(query_params.only, str):
        what_ = check_one(query_params.only)  # type: ignore

    query = f"select {what_} from {query_params.table_name}"
    if cond:
        query += f" {cond}"
    if query_params.order and isinstance(query_params.order, tuple):
        query += f" order by {parse_orders(query_params.order)}"
    if query_params.limit:
        query += f" limit {query_params.limit}"
    if query_params.offset:
        query += f" offset {query_params.offset}"

    return query, data


@lru_cache
def _build_update(query_params: QueryParams):
    check_one(query_params.table_name)
    cond, data = extract_signature(query_params.condition)
    new_str, updated = build_update_data(query_params.data)  # type: ignore
    query = f"update {query_params.table_name} set {new_str} {cond}"
    if query_params.limit:
        query = query.replace(cond, "")
        query += setup_limit_patch(query_params.table_name, cond, query_params.limit)
    if query_params.order:
        query += f" order by {parse_orders(query_params.order)}"
    # ? Require manual intervention to make sure updated is sync as
    # print(query)
    return query, data, updated
    # ? ... combine_keyvals(updated, NEW DATA)
    # ? our cache data only contain keys not values (v0.3.0)


@lru_cache
def _build_delete(query_params: QueryParams):
    check_one(query_params.table_name)
    cond, data = extract_signature(query_params.condition)
    query = f"delete from {query_params.table_name} {cond}"
    if query_params.limit:
        query = query.replace(cond, "")
        query += setup_limit_patch(query_params.table_name, cond, query_params.limit)
    if query_params.order:
        query += f" order by {parse_orders(query_params.order)}"
    return query, data


@lru_cache
def _build_insert(table_name: str, data: CacheData):
    check_one(table_name)
    converged = format_paramable(data)
    query = f"insert into {table_name} ({', '.join(val for val in converged)}) \
values ({', '.join(val for val in converged.values())})"
    return query, data


def build_update_data(data: dict[str, Any] | CacheData, suffix: str = "_set"):
    """Build update data, used to parameterized update data.
    Suffix is used to make sure there's no collisions with others. Use this with caution.
    """
    string = ""
    that: dict[str, str] = {}
    for key in data:
        check_one(key)
        string += f"{key}=:{key}{suffix}, "
        that[f"{key}{suffix}"] = f":{key}{suffix}"
    return string[:-2], that


def _handle_in(key, middle, val, condition_id):
    vals = tuple(
        f":prop_{condition_id}_val_in{index}" for index, _ in enumerate(val.data)
    )
    clause = f" {key} {middle} ({', '.join(vals)})"
    data = {key0[1:]: val0 for key0, val0 in zip(vals, val.data)}
    return clause, data


def _handle_between(key, middle, val):
    vdata = val.data
    if not all(isinstance(x, (int, float)) for x in vdata):
        raise SecurityError("Values for between constraint is not int/float")
    clause = f" {key} {middle} {vdata[0]!r} and {vdata[1]!r}"
    return clause


def _handle_like(key, middle, val):
    vdata = val.data
    clause = f" {key} {middle} {vdata!r}"
    return clause


def extract_signature(  # pylint: disable=too-many-locals
    filter_: Condition | CacheCond = None, suffix: str = "_check", depth: int = 0
):
    """Extract filter signature."""
    if depth < 0 or depth >= MAX_SUBQUERY_STACK_LIMIT:
        raise RecursionError(
            "Subquery builder has reached recursion limit of"
            f"{MAX_SUBQUERY_STACK_LIMIT}"
        )
    if filter_ is None:
        return "", {}

    if isinstance(filter_, (list, tuple)):
        filter_ = dict(filter_)

    call_id = generate_ids()
    clauses = []
    data: dict[str, Any] = {}

    for key, value in filter_.items():
        check_one(key)
        condition_id = generate_ids()
        if not isinstance(value, Signature):
            value = Signature(value, "=")
        old_data = value.value

        val = (
            Signature(
                ":"
                + NAMING_FORMAT.format(
                    key=key,
                    suffix=suffix,
                    call_id=call_id,
                    depth=depth,
                    condition_id=condition_id,
                ),
                value.generate(),
                value.data,
            )
            if value.value is not null
            else value
        )

        if isinstance(value.value, SubQuery):
            clause, subq_data = handle_subquery(key, value, depth)
            clauses.append(clause)
            data.update(subq_data)
            continue

        middle = val.generate()
        if val.normal_operator:
            clauses.append(f" {key}{middle}{val.value}")
        elif val.is_in:
            clause, in_data = _handle_in(key, middle, val, condition_id)
            clauses.append(clause)
            data.update(in_data)
            continue
        elif val.is_between:
            clause = _handle_between(key, middle, val)
            clauses.append(clause)
        elif val.is_like:
            clause = _handle_like(key, middle, val)
            clauses.append(clause)

        if val.value is not null:
            data[
                NAMING_FORMAT.format(
                    key=key,
                    suffix=suffix,
                    call_id=call_id,
                    depth=depth,
                    condition_id=condition_id,
                )
            ] = old_data

    if not clauses:
        return "", data

    where_clause = "where" + " and".join(clauses)
    return where_clause, data


@lru_cache
def extract_subquery(subquery: SubQuery, depth: int = 1):
    """Extract subquery into a valid SQL statement"""
    return _build_select(
        QueryParams(
            subquery.table,
            subquery.where,
            subquery.cols,
            subquery.limit,
            0,  # type: ignore
            subquery.orders,  # type: ignore
        ),
        depth=depth,
    )


def handle_subquery(key, value, depth):
    """Handle subquery data"""
    subq, subq_data = extract_subquery(value.value, depth=depth + 1)
    clause = f" {key} in ({subq})"
    return clause, subq_data



"""Type checking"""

from typing import TYPE_CHECKING, Any, Type
from dataclasses import fields, is_dataclass
from .helpers import validate

if TYPE_CHECKING:
    from . import BaseModel

def infer_type(name: str, type_: "Type[Any]"):
    """Infer type checking for specific columns"""
    if not isinstance(type_, type):
        return None

    @validate(name, f"{name} type is not {type_.__name__}") # type: ignore
    def function(instance: "Type[BaseModel]"):
        return isinstance(getattr(instance, name), type_)
    return function

def typecheck(cls: "Type[BaseModel]"):
    """Automatically pushed Runtime type checking"""
    if not is_dataclass(cls):
        raise TypeError(f"{cls.__name__} is not a dataclass")

    for field in fields(cls):
        name, type_ = field.name, field.type
        yield infer_type(name, type_) # type: ignore



"""Table Creation"""

from typing import Iterable, Any
from shlex import shlex

from .utils import filter_extraction, SQL_ACTIONS, DEFAULT_MAPPINGS

from ..typings import MasterQuery
from ..column import Column
from ..locals import SQLACTION, _SQLITETYPES

class TableCreationExtractor:
    """Base processor for table creation"""

    def __init__(self, columns: Iterable[Column], type_mappings: dict[str, str] | None):
        self.columns = columns
        self.type_mappings = _get_type_mappings(type_mappings)
        self.primaries: list[Column] = []
        self.foreigns: list[Column] = []
        self.string = ""

    def process_columns(self):
        """Process each columns"""
        self.string = _iterate_etbc_step1(
            self.columns, self.string, self.primaries, self.foreigns, self.type_mappings
        )

    def add_primary_keys(self):
        """Add primary key constraints"""
        if self.primaries:
            self.string += (
                f" primary key ({', '.join((col.name for col in self.primaries))}),"
            )

    def add_foreign_keys(self):
        """Add foreign key constraints"""
        for column in self.foreigns:
            stable, sname = column.source, column.source_column
            self.string += (
                f" foreign key ({column.name}) references {stable} ({sname})"
                f" on delete {_iterate_sql_action(column.on_delete)}"  # type: ignore
                f" on update {_iterate_sql_action(column.on_update)},"  # type: ignore
            )

    def extract(self) -> str:
        """Extract everything into a string"""
        self.process_columns()
        self.add_primary_keys()
        self.add_foreign_keys()
        return self.string[1:-1]

def _iterate_sql_action(action: SQLACTION):
    return SQL_ACTIONS.get(action, action)


def _get_type_mappings(type_mappings: dict[str, str] | None) -> dict[str, str]:
    """Get the merged type mappings."""
    maps = DEFAULT_MAPPINGS.copy()
    if type_mappings:
        maps.update(type_mappings)
    return maps


def extract_table_creations(
    columns: Iterable[Column], type_mappings: dict[str, str] | None = None
):
    """Extract columns classes to sqlite table creation query."""
    extractor = TableCreationExtractor(columns, type_mappings)
    return extractor.extract()


def _iterate_etbc_step1(
    columns: Iterable[Column],
    string: str,
    primaries: list[Column],
    foreigns: list[Column],
    maps: dict[str, str],
):
    for column in columns:
        ctype = maps.get(column.type, column.type)
        string += _process_column_constraints(column, ctype)
        if column.raw_source:
            foreigns.append(column)
        if column.primary:
            primaries.append(column)
        string += ","
    return string


def _process_column_constraints(column: Column, ctype: str) -> str:
    """Process constraints for a single column."""
    constraints = f" {column.name} {ctype}"
    if not column.nullable:
        constraints += " not null"
    if column.unique:
        constraints += " unique"
    if column.default:
        constraints += f" default {repr(column.default)}"
    return constraints

def extract_single_column(column: Column):
    """Extract a column class to sqlite column creation query"""
    foreign = column.foreign
    primary = column.primary
    string = f"{column.name} {column.type}"
    if not column.nullable:
        string += " not null"
    if column.unique:
        string += " unique"
    if column.default:
        string += f" default {repr(column.default)}"

    if primary:
        string += "primary key"
    if foreign:
        string += f"foreign key references {column.source} ({column.source_column})"
    return string

def basic_extract(table_creation: str):  # pylint: disable=too-many-locals
    """basic extraction for table"""
    data = table_creation[table_creation.find("(") + 1 : -1]
    cols = []
    upheld: dict[str, list[Any]] = {}
    for constr in data.split(","):
        name, type_, defaults, sources = "", "", None, ()
        primary = foreign = notnull = unique = False
        base_columns = list(shlex(constr))
        if base_columns[1] not in _SQLITETYPES:
            break  # Other constraint
        (
            name,
            type_,
        ) = (
            base_columns[0],
            base_columns[1],
        )
        if len(base_columns) == 2:
            # cols.append(Column(name, type_))  # type: ignore
            upheld[name] = [name, type_, False, None, False, False, True, None]
            continue
        for token in base_columns:
            token_lowered = token.lower()
            if token_lowered == "defaults":
                defaults = base_columns[base_columns.index(token) + 1]
                defaults = (
                    defaults
                    if defaults[0] != '"' or defaults[0] != "'"
                    else defaults[1:-1]
                )
            if token_lowered == "primary":
                primary = True
            if token_lowered == "foreign":
                foreign = True
            if token_lowered == "reference":
                tb_index = base_columns.index(token) + 1
                tb_col = tb_index + 1
                sources = (base_columns[tb_index], base_columns[tb_col][1:-1])
            if token_lowered == "null":
                notnull = base_columns[base_columns.index(token) - 1].lower() == "not"
            if token_lowered == "unique":
                unique = True
        upheld[name] = [
            name,
            type_,
            foreign,
            f"{sources[0]}/{sources[1]}" if sources else None,
            primary,
            unique,
            not notnull,
            defaults if defaults else None,
        ]
    return cols, upheld

def extract_table(  # pylint: disable=too-many-locals
    table_creation: str,
) -> list[Column]:
    """Extract SQLite table string"""
    data = table_creation[table_creation.find("(") + 1 : -1]
    cols, upheld = basic_extract(table_creation)
    shlexed = list(shlex(data))
    _, paren_wrap, filtered = filter_extraction(data, shlexed)

    # for efficiency, is this part efficient though?
    # although, can SOMEONE have a million columns on a sqlite table?
    # So, the efficiency on this blob isn't a concern.
    for column_string in filtered.split(","):
        column_shlexed = list(shlex(column_string))
        for tindex, token in enumerate(column_shlexed):
            if token.lower() == "primary":
                next_ = tindex + 2
                str_wrap = "".join(column_shlexed[next_ : next_ + 2])
                if str_wrap.startswith(":wrap"):
                    wrap = paren_wrap[str_wrap]
                else:
                    wrap = str_wrap
                for name in (
                    wrap[1:-1].split(",") if wrap.startswith("(") else wrap.split(",")
                ):
                    upheld[name][4] = True
                continue
            if token.lower() == "foreign":
                next_ = tindex + 2
                str_wrap = "".join(column_shlexed[next_ : next_ + 2])
                if str_wrap.startswith(":wrap"):
                    wrap = paren_wrap[str_wrap]
                else:
                    wrap = str_wrap
                name = (
                    wrap[1:-1] if wrap.startswith("'") or wrap.startswith('"') else wrap
                )
                name = wrap[1:-1] if wrap.startswith("(") else wrap
                tb_index = next_ + 3
                tb_col = tb_index + 1
                source_col_str = paren_wrap[f":{column_shlexed[tb_col+1]}"][1:-1]
                sources = f"{column_shlexed[tb_index]}/{source_col_str}"
                upheld[name][3] = sources
                upheld[name][2] = True
                if "delete" in column_shlexed:
                    delete_index = column_shlexed.index("delete")
                    upheld[name].append(column_shlexed[delete_index + 1])
                if "update" in column_shlexed:
                    uindex = column_shlexed.index("update")
                    upheld[name].append(column_shlexed[uindex + 1])

    for _, upheld_column in upheld.items():
        cols.append(Column(*upheld_column))
    return cols

def fetch_columns(_master_query: MasterQuery):
    """Fetch columns of a table. `master_query` is originated from select()
    on sqlite_master table"""
    sql = _master_query["sql"]
    return extract_table(sql)



"""Performance testing"""
# pylint: disable=all
from typing import Callable
from random import randint
from timeit import timeit
import os.path
from sys import path
path.insert(0, os.path.realpath(os.path.join(__file__, "../..")))

from luminadb import Column, Database, integer, text
from luminadb.signature import op
from luminadb.utils import crunch
from luminadb.functions import Function

count = Function("COUNT")

GROUP_BASE = [{
    "id": 0,
    "name": "root"
}, {
    "id": 1,
    "name": "user"
}]

GROUP_BASE_CRUNCHED = crunch(GROUP_BASE)

GROUP_NAMEBASE = [{
    "name": "root"
}, {
    "name": "user"
}]

USER_NAMEBASE = [{
    "username": 'root',
    'role': 'root'
}]

USER_BASE = [{
    "id": 0,
    "username": "root",
    "role": "root",
    "gid": 0
}, {
    "id": 1,
    "username": "user",
    "role": "user",
    "gid": 1
}]

GROUP_NEW = [{
    "id": 2,
    "name": "test"
}]

USER_NEW = [{
    "id": 2,
    "username": "test",
    "role": "test",
    "gid": 2
}]

GROUP_UNEW = [{
    "id": 3,
    "name": "test1"
}]

USER_UNEW = [{
    "id": 3,
    "username": "test0",
    "role": "test0",
}]

COUNT_DATA = [{
    "item_id": 1,
    "quantity": 50,
    "name": "A"
}, {
    "item_id": 2,
    "quantity": 50,
    "name": "A"

}, {
    "item_id": 3,
    "quantity": 50,
    "name": "A"

}, {
    "item_id": 4,
    "quantity": 50,
    "name": "A"

}]

performance_counter = {}


def setup_database(database: Database):
    """setup database"""
    users = database.create_table("users", [
        Column("id", "integer", unique=True, primary=True),
        Column("username", "text"),
        Column("role", "text", default="user"),
        Column("gid", "integer", foreign=True, foreign_ref="groups/id")
    ])
    groups = database.create_table("groups", [
        Column("id", "integer", primary=True),
        Column("name", "text")
    ])
    groups.insert_many(GROUP_BASE)
    users.insert_many(USER_BASE)


def setup_database_builder(database: Database):
    """Setup database with builder pattern for column"""
    users = database.create_table("users", [
        integer('id'),
        text('username'),
        text('role').default('user'),
        integer('gid').foreign('groups/id')
    ])
    groups = database.create_table("groups", [
        integer('id'),
        text('name')
    ])
    groups.insert_many(GROUP_BASE)
    users.insert_many(USER_BASE)
    database.commit()

def setup_database_fns(database: Database):
    """Setup database used for functions"""
    checkout = database.create_table("checkout", [
        integer("item_id"),
        text("name"),
        integer('quantity')
    ])
    checkout.insert_many(COUNT_DATA)
    database.commit()

def setup_database_1mdata(database: Database):
    """Setup database and fill it with 1m data"""
    items = database.create_table('items', [
        text('name'),
        integer('quantity')
    ])
    items.insert_many([{'name': f'item {randint(0, 1_000_000)}', 'quantity': randint(0, 100)} for _ in range(1_000_000)])
    database.commit()

def init_memdb(initializer: Callable[[Database], None]):
    db = Database(":memory:")
    initializer(db)
    return db

def testp_00_select():
    """Test Performance 0000 select"""
    db = init_memdb(setup_database_builder)
    users = db.table("users")
    performance_counter['generic_select:all'] = timeit(lambda: users.select())
    assert True

def testp_00_01_select_one():
    """Test Performance 0001 select one"""
    db = init_memdb(setup_database_builder)
    users = db.table("users")
    performance_counter['specific_select:one'] = timeit(lambda: users.select_one())
    assert True

def testp_00_02_select_only():
    """Test Performance 0003 select only"""
    db = init_memdb(setup_database_builder)
    users = db.table("users")
    performance_counter['specific_select:only_one'] = timeit(lambda: users.select_one({'id': 0}, ('username', 'role')))
    assert True

def testp_00_03_select_crunch():
    """Test Performance 0004 select with crunch"""
    db = init_memdb(setup_database_builder)
    groups = db.table('groups')
    performance_counter['generic_select:crunched_return'] = timeit(lambda: groups.select(squash=True))
    assert True

def testp_01_insert():
    """test 0100 insert"""
    db = init_memdb(setup_database_builder)
    groups = db.table('groups')
    performance_counter['generic_insert'] = timeit(lambda: groups.insert_many(GROUP_NEW))
    groups.delete()

def testp_02_01_update():
    """test 0201 update"""
    db = init_memdb(setup_database_builder)
    groups = db.table('groups')
    performance_counter['generic_update'] = timeit(lambda: (groups.update({"id": op == 1}, GROUP_UNEW[0]), groups.update({'id': 3}, {'id': 1})), number=500_000)

def testp_03_00_delete():
    """Test Performance 0300 single deletion test"""
    db = init_memdb(setup_database_1mdata)
    items = db.table('items')
    performance_counter['generic_delete'] = timeit(lambda: items.delete_one())

def testp_03_01_delete_all():
    """Test 0301 deletion test (all)"""
    db = init_memdb(setup_database_1mdata)
    items = db.table('items')
    performance_counter['generic_delete_all'] = timeit(lambda: items.delete(), number=1)

def testp_99_99_final():
    """Test final"""
    with open("perf-counter.txt", 'w') as f:
        sto = '\n'.join(f'{key} -> {value}' for key, value in performance_counter.items())
        f.write(f"""Performance testing:\n{sto}\n{'='*30}""")



"""Indexes"""

from json import dumps
from .utils import check_one, check_iter

class Index:
    """Index"""

    def __init__(
        self,
        name: str = "",
        target: str = "",
        columns: tuple[str] | None = None,
        unique: bool = False,
    ) -> None:
        self._name = name
        self._target = target
        self._columns = columns if columns is not None else ()
        self._unique = unique

    def name(self, name: str):
        """Set this index name"""
        self._name = name
        return self

    def target(self, target: str):
        """Set this index target table"""
        self._target = target
        return self

    def columns(self, *columns: str):
        """Set which index this index is based on"""
        self._columns = columns
        return self

    def unique(self):
        """Set whether this index is unique"""
        self._unique = True
        return self

    def build_sql(self):
        """Return current information as SQL"""
        if not all(map(bool, (self._name, self._target, self._columns))):
            raise ValueError("Either name, target, or column is empty.")
        check_one(self._name)
        check_one(self._target)
        check_iter(self._columns)
        unique = "unique" if self._unique else ""
        columns = ", ".join(self._columns)
        return f"create {unique} index {self._name} on {self._target} ({columns})"

    @property
    def index_name(self):
        """Return index name"""
        return self._name

    @property
    def index_target(self):
        """Return index target"""
        return self._target

    @property
    def index_columns(self):
        """Return columns target"""
        return self._columns

    @property
    def index_unique(self):
        """Return unique property"""
        return self._unique

    def to_json(self):
        """Return current information as JSON"""
        return dumps(
            {
                "name": self._name,
                "target": self._target,
                "columns": self._columns,
                "unique": self._unique,
            }
        )



"""Table API insertion tests"""

from pytest import raises
from luminadb import Database, text, SKIP
from luminadb.errors import CuteDemonLordException

from ..setup import groups, users, database, save_report, GROUP_NEW, USER_NEW

def test_insert():
    """test 0100 insert"""
    assert groups.insert_many(GROUP_NEW) is None
    assert users.insert_many(USER_NEW) is None
    users._sql.commit()  # pylint: disable=protected-access
    assert not not users.select()  # pylint: disable=unnecessary-negation
    assert save_report("01_insert", database, groups, users)


def test_insert_with():
    """Test 0101 insert in with-statement"""
    db = Database(":memory:")
    table = db.create_table("a", [text("name")])
    try:
        with table:
            table.insert({"name": "Admin"})
            raise CuteDemonLordException(
                "Haha, i'm cute demon lord and I cast [Insert Dispell]"
            )
    except CuteDemonLordException:
        pass
    assert table.select() == []


def test_insert_empty():
    """Test 0101 insert in with-statement"""
    db = Database(":memory:")
    table = db.create_table("a", [text("name")])
    with raises(ValueError):
        table.insert({"name": SKIP})



"""Custom operators for shorthand."""

from typing import Any
from .signature import Signature, op


def eq(name: str, other: Any) -> tuple[str, Signature]:
    """Same as using op == other"""
    return (name, op == other)


def lt(name: str, other: Any) -> tuple[str, Signature]:
    """Same as using op < other"""
    return (name, op < other)


def le(name: str, other: Any) -> tuple[str, Signature]:
    """Same as using op <= other"""
    return (name, op <= other)


def gt(name: str, other: Any) -> tuple[str, Signature]:
    """Same as using op > other"""
    return (name, op > other)


def ge(name: str, other: Any) -> tuple[str, Signature]:
    """Same as using op >= other"""
    return (name, op >= other)


def ne(name: str, other: Any) -> tuple[str, Signature]:
    """Same as using op != other"""
    return (name, op != other)


def like(name: str, condition: str) -> tuple[str, Signature]:
    """Like constraint"""
    return (name, op.like(condition))


def between(name: str, low: int, high: int) -> tuple[str, Signature]:
    """Between constraint"""
    return (name, op.between(low, high))

def in_(values: list[Any]):
    """In VALUES"""
    return this.in_(values)


this = op

__all__ = ["eq", "lt", "le", "gt", "ge", "ne", "like", "between", "this", 'in_']



"""Core builder"""

from typing import Any, Optional, Literal

from .typings import Condition
from .engine import QueryParams, _build_delete, _build_insert, _build_select, _build_update
from .utils import (
    setup_hashable,
    remove_null,
    combine_keyvals,
)

from ..typings import Orders, Data
from ..functions import ParsedFn # type: ignore

def build_select(  # pylint: disable=too-many-arguments
    table_name: str,
    condition: Condition = None,
    only: tuple[str, ...] | ParsedFn | str | Literal["*"] = "*",
    limit: int = 0,
    offset: int = 0,
    order: Optional[Orders] = None,
) -> tuple[str, dict[str, Any]]:
    """Build select query (this function (backendly) cache!)

    Args:
        table_name (str): Table name
        condition (Condition, optional): Condition to use. Defaults to None.
        only: (OnlyColumn, optional): Select what you want. Default to None.
        limit (int, optional): Limit query (this also limits DB-API 2 `.fetchall`). Defaults to 0.
        offset (int, optional): Offset. Defaults to 0.
        order (Optional[Orders], optional): Order. Defaults to None.

    Returns:
        tuple[str, dict[str, Any]]: query and query data
    """
    cond, order_, _ = setup_hashable(condition, order)
    params = QueryParams(
        table_name=table_name,
        condition=cond,
        only=only,
        limit=limit,
        offset=offset,
        order=order_,  # type: ignore
    )
    return _build_select(params)

def build_update(
    table_name: str,
    new_data: Data,
    condition: Condition = None,
    limit: int = 0,
    order: Optional[Orders] = None,
) -> tuple[str, dict[str, Any]]:
    """Build update query (once again, this function backendly cache)

    Args:
        table_name (str): Table name
        new_data (Data): New data to update
        condition (Condition, optional): Condition to limit what to update. Defaults to None.
        limit (int, optional): limit chanes. Defaults to 0.
        order (Optional[Orders], optional): What order of change?. Defaults to None.

    Returns:
        tuple[str, dict[str, Any]]: query, query data
    """
    new_data = remove_null(new_data)
    cond, order_, ndata = setup_hashable(condition, order, new_data)
    params = QueryParams(
        table_name=table_name,
        condition=cond,
        limit=limit,
        order=order_,  # type: ignore
        data=ndata,
    )
    query, check, updated = _build_update(params)
    return query, check | combine_keyvals(updated, new_data)


def build_delete(
    table_name: str,
    condition: Condition = None,
    limit: int = 0,
    order: Optional[Orders] = None,
) -> tuple[str, dict[str, Any]]:
    """Build delete query

    Args:
        table_name (str): Table name
        condition (Condition, optional): Condition to limit deletion. Defaults to None.
        limit (int, optional): Limit to limit deletion. Defaults to 0.
        order (Optional[Orders], optional): Order. Defaults to None.

    Returns:
        tuple[str, dict[str, Any]]: query, query data
    """

    cond, order_, _ = setup_hashable(condition, order)
    params = QueryParams(
        table_name, condition=cond, limit=limit, order=order_  # type: ignore
    )
    return _build_delete(params)


def build_insert(table_name: str, data: Data) -> tuple[str, dict[str, Any]]:
    """Build insert query

    Args:
        table_name (str): table name
        data (Data): Data to insert

    Returns:
        tuple[str, dict[str, Any]]: query, query data
    """
    data = remove_null(data)
    _, _, ndata = setup_hashable(None, None, data)
    query, _ = _build_insert(table_name, ndata)
    return query, data



"""Setup"""

# pylint: disable=redefined-outer-name,invalid-name
from json import dumps
from io import StringIO
from os.path import abspath
from types import SimpleNamespace
from tempfile import mkdtemp
from pathlib import Path

from luminadb import Column, Database, integer, text
from luminadb.models import Primary, Unique, model, BaseModel, Foreign, CASCADE
from luminadb.utils import crunch
from luminadb.functions import Function

def parse(data):
    """parse"""
    return dumps(data, indent=2)


temp_dir = Path(mkdtemp())

config = SimpleNamespace()
file = abspath(f"{__file__}/../../reports.txt")
pstdout = StringIO()

count = Function("COUNT")

GROUP_BASE = [{"id": 0, "name": "root"}, {"id": 1, "name": "user"}]

GROUP_BASE_CRUNCHED = crunch(GROUP_BASE)

GROUP_NAMEBASE = [{"name": "root"}, {"name": "user"}]

USER_NAMEBASE = [{"username": "root", "role": "root"}]

USER_BASE = [
    {"id": 0, "username": "root", "role": "root", "gid": 0},
    {"id": 1, "username": "user", "role": "user", "gid": 1},
]

GROUP_NEW = [{"id": 2, "name": "test"}]

USER_NEW = [{"id": 2, "username": "test", "role": "test", "gid": 2}]

GROUP_UNEW = [{"id": 3, "name": "test1"}]

USER_UNEW = [
    {
        "id": 3,
        "username": "test0",
        "role": "test0",
    }
]

COUNT_DATA = [
    {"item_id": 1, "quantity": 50, "name": "A"},
    {"item_id": 2, "quantity": 50, "name": "A"},
    {"item_id": 3, "quantity": 50, "name": "A"},
    {"item_id": 4, "quantity": 50, "name": "A"},
]


def setup_database(database: Database):
    """setup database"""
    users = database.create_table(
        "users",
        [
            Column("id", "integer", unique=True, primary=True),
            Column("username", "text"),
            Column("role", "text", default="user"),
            Column("gid", "integer", foreign=True, foreign_ref="groups/id"),
        ],
    )
    groups = database.create_table(
        "groups", [Column("id", "integer", primary=True), Column("name", "text")]
    )
    groups.insert_many(GROUP_BASE)
    users.insert_many(USER_BASE)


def setup_database_builder(database: Database):
    """Setup database with builder pattern for column"""
    users = database.create_table(
        "users",
        [
            integer("id").primary(),
            text("username"),
            text("role").default("user"),
            integer("gid").foreign("groups/id"),
        ],
    )
    groups = database.create_table("groups", [integer("id").primary(), text("name")])
    groups.insert_many(GROUP_BASE)
    users.insert_many(USER_BASE)
    database.commit()


def setup_database_fns(database: Database):
    """Setup database used for functions"""
    checkout = database.create_table(
        "checkout", [integer("item_id").primary(), text("name"), integer("quantity")]
    )
    checkout.insert_many(COUNT_DATA)
    database.commit()


def setup_orderable(database: Database):
    """Setup database with order-able data"""
    items = database.create_table("items", [text("name"), integer("quantity")])
    items.insert_many([{"name": "a", "quantity": a} for a in range(100)])
    database.commit()


def setup_model_api(database: Database):
    # pylint: disable=missing-class-docstring
    """Setup Model API Users/Posts"""

    @model(database)
    class Users(BaseModel):
        __schema__ = (Primary("id"), Unique("username"))
        id: str
        username: str
        is_active: bool = True

    @model(database)
    class Posts(BaseModel):
        __schema__ = (Primary("id"),
                      Foreign("user_id", Users)
                              .on_delete(CASCADE)
                              .on_update(CASCADE))
        id: str
        user_id: str
        title: str
        content: str

    return Users, Posts

database = Database(temp_dir / "test.db")  # type: ignore
setup_database_builder(database)
setup_database_fns(database)
users = database.table("users")
groups = database.table("groups")

def save_report(tid, database, grouptb, usertb):
    """save report"""
    master = database.table("sqlite_master")
    print(
        f">> {tid}",
        "\n>> master ",
        parse(master.select()),
        "\n>> groups",
        parse(grouptb.select()),
        "\n>> users",
        parse(usertb.select()),
        end="\n=======\n",
        file=pstdout,
    )
    return True



"""Models"""

# pylint: disable=unused-import,unused-argument,cyclic-import,protected-access

from contextlib import contextmanager
from typing import Any, Callable, Self, Type, TypeVar
from dataclasses import asdict, dataclass, fields, is_dataclass, MISSING

from luminadb.models.type_checkers import typecheck

from .helpers import (
    Constraint,
    Unique,
    Primary,
    Foreign,
    TYPES,
    Validators,
    CASCADE,
    DEFAULT,
    NOACT,
    RESTRICT,
    SETNULL,
)

from .helpers import VALID_HOOKS_NAME, hook, validate, initiate_hook, initiate_validators
from .query_builder import QueryBuilder
from .errors import ConstraintError
from ..errors import DatabaseExistsError
from ..database import Database, Table
from ..column import text, BuilderColumn
from ..operators import in_

NULL = object()
T = TypeVar("T", bound="BaseModel")

## Model functions


@staticmethod
def noop_autoid():
    """Default no-op function for BaseModel __auto_id__"""
    return None


class BaseModel:  # pylint: disable=too-few-public-methods,too-many-public-methods
    """Base class for all Models using Model API"""

    __table_name__ = ""
    __schema__: tuple[Constraint, ...] = ()
    __validators__: dict[str, list[Validators]] = {}
    __hooks__: "dict[str, list[Callable[[Self], None] | str]]" = {}
    __hidden__: tuple[str, ...] = ()
    __auto_id__: Callable[[], Any] = noop_autoid
    _tbl: Table
    _primary: str | None

    @classmethod
    def create_table(cls, db: Database):
        """Create table according to annotations and schema from `__schema__`"""
        if not is_dataclass(cls):
            raise TypeError(f"{cls.__name__} must be a dataclass")

        columns: list[BuilderColumn] = []
        constraints: dict[str, list[Constraint]] = {
            col: [] for col in cls.__annotations__
        }  # pylint: disable=no-member
        cls.__table_name__ = cls.__table_name__ or cls.__name__.lower()
        _primary = None

        # Extract constraints from __schema__
        for constraint in cls.__schema__:  # pylint: disable=no-member
            target_col = constraint.column
            constraints[target_col].append(constraint)
            if isinstance(constraint, Primary) and _primary:
                raise ConstraintError("Cannot apply when a column has 2 primary keys")
            if isinstance(constraint, Primary):
                _primary = target_col
                continue

        # Process fields & constraints
        for field_def in fields(cls):  # Fetch fields using dataclass reflection
            field_name = field_def.name
            field_type = field_def.type
            col = TYPES.get(field_type, text)(field_name)  # type: ignore

            # Check if there's a default value
            if field_def.default is not MISSING:
                default_value = field_def.default
                col = col.default(default_value)

            # Apply constraints dynamically
            for constraint in constraints.get(field_name, []):
                constraint.apply(col)

            columns.append(col)

        cls._primary = _primary
        try:
            cls._tbl = db.create_table(cls.__table_name__, columns)
        except DatabaseExistsError:
            cls._tbl = db.table(cls.__table_name__.lower())

    @classmethod
    def _execute_hooks(cls, name: str, instance: Self):
        for hook_fn in cls.__hooks__.get(name, ()):
            if isinstance(hook_fn, str):
                getattr(cls, hook_fn)(instance)
            else:
                hook_fn(instance)

    @classmethod
    def _execute_validators(cls, name: str, instance: Self):
        for validator_fn in cls.__validators__.get(name, ()):
            validator_fn.validate(instance)

    @classmethod
    def _register(cls, type_: str = "hook", name: str = "", if_fail: str = ""):
        """Register a hook/validator under a name"""
        if type_ not in ("hook", "validator"):
            raise ValueError("Which do you want?")

        def function(func):
            if name == "":
                raise ValueError(f"{type_.title()} name needs to be declared.")
            if type_ == "hook":
                if name not in VALID_HOOKS_NAME:
                    raise ValueError("Name of a hook doesn't match with expected value")
                cls.__hooks__.setdefault(name, [])
                if cls.__hooks__[name]:
                    cls.__hooks__[name] = [func]
                else:
                    cls.__hooks__[name].append(func)
                return func

            if not is_dataclass(cls):
                raise TypeError("Dataclass is required for this class")

            fields_ = tuple((field.name for field in fields(cls)))
            if name not in fields_:
                raise ValueError(f"Expected validator to has name as column field. Got {name!r}")

            fail = if_fail or f"{name} fails certain validator"

            cls.__validators__.setdefault(name, [])
            validator_entry = Validators(func, fail)
            if cls.__validators__[name]:
                cls.__validators__[name] = [validator_entry]
            else:
                cls.__validators__[name].append(validator_entry)
            return func

        return function

    @classmethod
    def create(cls, **kwargs):
        """Create data based on kwargs"""
        primary: str | None = cls._primary or kwargs.get("id", None)
        id_present = bool(kwargs.get(cls._primary or "id", None))
        if primary and cls.__auto_id__ and not id_present:  # type: ignore
            kwargs[primary] = cls.__auto_id__()  # type: ignore
        instance = cls(**kwargs)

        cls._execute_hooks("before_create", instance)
        for key in kwargs:
            cls._execute_validators(key, instance)
        cls._tbl.insert(kwargs)
        cls._execute_hooks("after_create", instance)
        return instance

    def update(self, __primary: str | object = NULL, /, **kwargs):
        """Update current data"""
        # pylint: disable=protected-access
        primary = self._primary or __primary
        if primary is NULL:
            raise ValueError(
                "The table does not have any primary key, cannot update due to undefined selection"
            )
        self._execute_hooks("before_update", self)
        for key in kwargs:
            self._execute_validators(key, self)
        self._tbl.update({primary: getattr(self, primary)}, kwargs)  # type: ignore
        for key, value in kwargs.items():
            setattr(self, key, value)
        self._execute_hooks("after_update", self)
        return self

    def delete(self, __primary=NULL, /):
        """Delete current data"""
        # pylint: disable=protected-access
        primary = self._primary or __primary
        if primary is NULL:
            raise ValueError(
                "The table does not have any primary key, cannot delete due to undefined selection"
            )
        self._execute_hooks("before_delete", self)
        self._tbl.delete_one({primary: getattr(self, primary)})  # type: ignore
        self._execute_hooks("after_delete", self)

    @classmethod
    def bulk_create(cls, records: list[dict]):
        """Insert multiple records at once."""
        cls._tbl.insert_many(records)
        return [cls(**record) for record in records]  # Return list of instances

    @classmethod
    def bulk_update(cls, records: list[dict], key: str | object = NULL):
        """Update multiple records using a primary key or provided key."""
        key_ = cls._primary or key
        if key is NULL:
            raise ValueError(
                "The table does not have any primary key, or key parameter is not provided"
            )
        for record in records:
            if key_ not in record:
                raise ValueError(f"Missing primary key '{key_}' in record: {record}")
            cls._tbl.update({key_: record[key_]}, record)  # type: ignore

    @classmethod
    def bulk_delete(cls, keys: list[Any], key: str):
        """Delete multiple records using a primary key."""
        cls._tbl.delete({key: in_(keys)})

    @classmethod
    def first_or_fail(cls, **kwargs):
        """Return the first matching record or raise an error if no match is found."""
        result = cls.where(**kwargs).limit(1).throw().fetch_one()
        return result

    @classmethod
    def first(cls, **kwargs):
        """Return the first matching record or None if no match is found."""
        result = cls.where(**kwargs).limit(1).fetch_one()
        return result

    @classmethod
    def one(cls, **kwargs):
        """Return exactly one record. Raises error if multiple results exist."""
        results = cls.where(**kwargs).fetch()
        if len(results) > 1:
            raise ValueError(f"Expected one record, but found {len(results)}")
        return results[0] if results else None

    @classmethod
    def find(cls, amount: int):
        """Return models relative to the amount"""
        results = cls.query().limit(amount).fetch()
        return results

    @classmethod
    def find_or_fail(cls, amount: int):
        """Return models relative to the amount and when returned is equal to 0, throws an error"""
        results = cls.query().limit(amount).throw().fetch()
        return results

    @classmethod
    def all(cls):
        """Return all values from the table"""
        return cls.where().fetch()

    @classmethod
    def count(cls, **kwargs) -> int:
        """Return count of matching records."""
        return cls.where(**kwargs).count()

    @classmethod
    def exists(cls, **kwargs) -> bool:
        """Check if any record matches the query."""
        return cls.where(**kwargs).limit(1).fetch_one() is not None

    @classmethod
    @contextmanager
    def atomic(cls):
        """Perform operations within a transaction."""
        with cls._tbl:  # Assuming `transaction()` exists
            yield

    @classmethod
    def upsert(cls, key: str, **kwargs):
        """Insert or update a record based on primary key."""
        existing = cls.where(**{key: kwargs[key]}).fetch_one()
        if existing:
            return existing.update(**kwargs)
        return cls.create(**kwargs)

    def to_dict(self):
        """Convert model instance to dictionary."""
        if is_dataclass(self):  # always true, though, just in case
            instance = asdict(self)
            return {k: v for k, v in instance.items() if k not in self.__hidden__}
        return {}

    def to_safe_instance(self) -> Self:
        """Wrap instance that complies with __hidden__."""
        if is_dataclass(self):
            dict_inst = asdict(self).items()
            instance = {
                k: (v if not k in self.__hidden__ else None) for k, v in dict_inst
            }
            return type(self)(**instance)
        raise TypeError("This class must be a dataclass")

    def raw(self, query: str, params: list[Any] | tuple[Any, ...] | dict[str, Any]):
        """Raw SQL query"""
        return self._tbl._sql.execute(query, params)  # pylint: disable=protected-access

    def has_many(self, related: "Type[T]", foreign_key: str | object = NULL):
        """Ensure related_model has a Foreign key pointing to self"""
        foreign_key = None

        if not self._primary:
            raise ConstraintError(
                f"The table {self.__table_name__} does not have any primary key "
                "required for has_many()"
            )

        # Scan __schema__ of related model to find a Foreign key linking back
        for constraint in related.__schema__:
            if isinstance(constraint, Foreign):
                table_ref, _ = constraint.target.split("/")  # type: ignore
                if table_ref == self.__class__.__name__.lower():
                    foreign_key = constraint.column
                    break

        if not foreign_key:
            raise ValueError(
                f"{related.__name__} does not have a Foreign key pointing "
                f"to {self.__class__.__name__}"
            )

        # Perform the actual query
        return related.where(**{foreign_key: getattr(self, self._primary)}).fetch()

    def belongs_to(self, related_model: "Type[T]"):
        """Retrieve the related model that this instance belongs to."""
        # Find the Foreign() constraint that references `related_model`
        for constraint in self.__schema__:
            if isinstance(constraint, Foreign) and constraint.target.startswith(  # type: ignore
                related_model.__table_name__ + "/"
            ):
                foreign_key = constraint.column
                referenced_column = constraint.target.split("/")[1]  # type: ignore
                return related_model.where(
                    **{referenced_column: getattr(self, foreign_key)}
                ).fetch_one()

        raise ValueError(
            f"{self.__class__.__name__} does not belong to {related_model.__name__}"
        )

    def has_one(self, related_model: "Type[T]"):
        """Retrieve the related model where this instance is referenced."""
        # Find the Foreign() constraint in `related_model` that references this model
        for constraint in related_model.__schema__:
            if (
                isinstance(constraint, Foreign)
                and constraint.target == f"{self.__table_name__}/{self._primary}"
            ):
                foreign_key = constraint.column
                return related_model.where(**{foreign_key: self._primary}).fetch_one()

        raise ValueError(
            f"{related_model.__name__} does not have a one-to-one relationship "
            f"with {self.__class__.__name__}"
        )

    @classmethod
    def get_table(cls):
        """Return table instance"""
        return cls._tbl

    @classmethod
    def where(cls, **kwargs):
        """Basic select operation"""
        return QueryBuilder(cls).where(**kwargs)

    @classmethod
    def query(cls):
        """Return Query Builder related to this model"""
        return QueryBuilder(cls)

def model(db: Database, type_checking: bool = False):
    """Initiate Model API compatible classes. Requires target to be a dataclass,
    the app automatically injects dataclass if this isn't a dataclass.

    Use `type_checking` if you want automatic runtime type checker."""

    def outer(cls: Type[T]) -> Type[T]:
        if not issubclass(cls, BaseModel):
            raise TypeError(f"Model {cls.__name__} is not subclass of BaseModel.")
        if not is_dataclass(cls):
            cls = dataclass(cls)
        cls.create_table(db)
        if type_checking:
            for fn in typecheck(cls):
                initiate_validators(cls, fn)

        for member in cls.__dict__.values():
            initiate_hook(cls, member)
            initiate_validators(cls, member)
        return cls

    return outer


__all__ = [
    "model",
    "BaseModel",
    "Unique",
    "Primary",
    "Foreign",
    "QueryBuilder",
    "CASCADE",
    "DEFAULT",
    "NOACT",
    "SETNULL",
    "RESTRICT",
    "validate",
    "hook"
]



from time import perf_counter
from typing import Any, Callable
from functools import wraps
from typing import NamedTuple

class Result(NamedTuple):
    result: Any
    time: float

def timed(func: Callable):
    """Decorate functions that added performance timing. This is only timed the function once per-call.
    
    Returns: Result"""
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = perf_counter()
        result = func(*args, **kwargs)
        end = perf_counter()
        return Result(result, end-start)
    return wrapper

def timed_oncall(func: Callable, *args, **kwargs):
    """Immediately calls a function and return Result."""
    start = perf_counter()
    result = func(*args, **kwargs)
    end = perf_counter()
    return Result(result, end-start)



"""Used for debugging"""

from inspect import getouterframes, FrameInfo, Traceback

STATE = {"DEBUG": False}


def _build_frame_str(frame: FrameInfo | Traceback):
    context_lines = (
        "<no context>\n"
        if not frame.code_context
        else "\n>>> ".join(frame.code_context)
    )
    lineno = frame.lineno or "<None>"
    filename = frame.filename or "<None>"
    function = frame.function or "<Callable()>"
    return f"""\
Trace at {lineno}, file {filename}, {function}
>>> {context_lines}"""


def map_frame(frame: Traceback) -> str:
    """Map a Taceback or FrameInfo to readable trace"""
    allframes = getouterframes(frame)
    allframes.reverse()

    builded = [_build_frame_str(frame) for frame in allframes]
    return "".join(builded)


def map_frames(frames: list[FrameInfo | Traceback]):
    """Shorthand for alreadby built list of frames"""
    return "".join((_build_frame_str(frame) for frame in frames))


def if_debug_print(*args, sep=" ", end="\n", flush=True):
    """If debug? print!"""
    if STATE["DEBUG"]:
        arg0 = args[0]
        print(
            arg0,
            *(repr(arg) if arg != "\n" else "\n" for arg in args if args.index(arg) != 0),
            sep=sep,
            end=end,
            flush=flush,
        )



# pylint: disable=all
from sys import argv
from json import loads
from io import StringIO
from typing import NamedTuple

buffer = StringIO()

class Entry(NamedTuple):
    type: str
    symbol: str
    message: str
    messageId: str
    confidence: str
    module: str
    obj: str
    line: int
    column: int
    endLine: int
    endColumn: int
    path: str
    absolutePath: str

if not len(argv) > 1:
    print("Requiring pylint json output!")

with open(argv[1]) as outfile:
    data = loads(outfile.read())

def write_source(entry: Entry):
    with open(entry.path) as fstream:
        dt = [a.rstrip() for a in fstream.readlines()]
    if entry.line == entry.endLine:
        selected = dt[entry.line]
    else:
        selected = dt[entry.line-1:entry.endLine+1]
    buffer.write(f"""\
[{entry.module}] {entry.message} ({entry.symbol})
{''.join(selected)}\n\n""")

stats = data["statistics"]
msg_types = stats["messageTypeCount"]
msgs = data["messages"]

buffer.write(f"""\
[pylint] Summary
Score  = {stats['score']}
Linted = {stats['modulesLinted']}
---
Fatal      = {msg_types['fatal']}
Error      = {msg_types['error']}
Warning    = {msg_types['warning']}
Refactor   = {msg_types['refactor']}
Convention = {msg_types['convention']}
Info       = {msg_types['info']}

{' Records ':-^20}
""")

for i in data["messages"]:
    write_source(Entry(**i))

buffer.write('-'*20)

with open("records/pylint.txt", "w") as record:
    record.write(buffer.getvalue())

print(f"[pylint] Project score is {data['statistics']['score']}. See the report in records/pylint.txt")



"""Signature"""

from typing import Any, Optional

from .utils import matches, null
from .errors import SecurityError
from .locals import _NO_UNLIKE
from .typings import tuple_list


class Signature:
    """Signature

    This class is used at most filter parameters at `Table` class.

    You can use variable op (operator) to do all conditional logic or operator.
    There are ==, <=, <, !=, >, and >=

    .negate() adds 'not' to the operator, so from == would equal to != when used with
    .negate()

    .like() only search for alike characters. Please refer to sqlite like command.

    .between() is a ranging operator, there's low and high value.

    Example of everything:
        >>> op == 5
        >>> op <= 2
        >>> op < 2
        >>> op != 5
        >>> op > 7
        >>> op >= 7
        >>> op.like('Ar%')
        >>> op.like('Ar_')
        >>>op.between(1, 10)

    You can now use function from :operators:"""

    def __init__(
        self,
        value: Any = null,
        operator: Optional[str] = None,
        data: Optional[tuple[int, int] | str] | tuple_list[Any] = None,
        negate=False,
    ) -> None:
        self._value = value
        self._operator: str = "" if operator is None else operator
        self._data = data
        self._negate = negate

    def __eq__(self, __o) -> "Signature":
        return Signature(__o, "=")

    def __lt__(self, __o) -> "Signature":
        return Signature(__o, "<")

    def __le__(self, __o) -> "Signature":
        return Signature(__o, "<=")

    def __gt__(self, __o) -> "Signature":
        return Signature(__o, ">")

    def __ge__(self, __o) -> "Signature":
        return Signature(__o, ">=")

    def __ne__(self, __o) -> "Signature":
        return Signature(__o, "!=")

    def like(self, str_condition: str):
        """Like"""
        is_valid = not matches(_NO_UNLIKE, str_condition)
        if not is_valid:
            raise SecurityError("Cannot understand other character.")
        return Signature(null, "like", str_condition, self._negate)

    def in_(self, values: list[Any]):
        """IN"""
        return Signature(null, operator="in", data=tuple(values), negate=self._negate)

    def between(self, low: int, high: int):
        """Betweeen"""
        return Signature(null, "between", (low, high), self._negate)

    def negate(self):
        """Negate or adding NOT"""
        return Signature(
            self._value, self._operator, tuple(self._data), not self._negate  # type: ignore
        )

    @property
    def value(self):
        """value"""
        return self._value

    @property
    def data(self):
        """data"""
        return self._data

    @property
    def kind_sign(self):
        """Signature kind"""
        return {
            "=": "eq",
            "<": "lt",
            "<=": "le",
            ">": "gt",
            ">=": "ge",
            "!=": "ne",
        }.get(self._operator, self._operator)

    @property
    def normal_operator(self):
        """Is it normal?"""
        return self._operator in ("!=", "=", "<", "<=", ">", ">=")

    @property
    def negated(self):
        """Is negated"""
        return self._negate

    @property
    def is_between(self):
        """Is operator between"""
        return self._operator == "between"

    @property
    def is_like(self):
        """Is operator like"""
        return self._operator == "like"

    @property
    def is_in(self):
        """Is operator in?"""
        return self._operator == "in"

    def generate(self):
        """Generate operator string"""
        string = "not " if self.negated else ""
        string += self._operator
        return string

    def __hash__(self):
        return hash((self._value, self._data, self._negate, self._operator))

    def __repr__(self) -> str:
        return f"<Signature -> {self._operator} {self._data!r}>"


op = Signature()



"""CSV module, used to export database/table to csv"""

from csv import DictWriter, DictReader
from io import StringIO
from os import mkdir
from os.path import join as join_path, isfile, exists

from . import Database, Table


def _export_table(table: Table) -> str:
    data = StringIO()
    tb_data = table.select_one()
    if tb_data is None:
        return ""
    fields = tuple(tb_data.keys())
    csv_writer = DictWriter(data, fields)
    csv_writer.writeheader()
    for selecteds in table.paginate_select(length=1):
        if len(selecteds) == 0:
            break
        selected = selecteds[0]
        csv_writer.writerow(selected)
    return data.getvalue()


def _export_database(database: Database) -> list[tuple[str, str]]:
    exported = []
    for table in database.tables():
        exported.append((table.name, _export_table(table)))
    return exported


def _process_database(filename: str, data: list[tuple[str, str]]):
    for table in data:
        base = join_path(filename, f"{table[0]}.csv")
        with open(base, "w", encoding="utf-8") as file:
            file.write(table[1])


def to_csv_string(table_or_database: Table | Database):
    """Export database/table to csv"""
    if isinstance(table_or_database, Table):
        return _export_table(table_or_database)
    if isinstance(table_or_database, Database):
        return _export_database(table_or_database)
    raise TypeError(
        f"Expected Table or Database, got {type(table_or_database).__name__}"
    )


def to_csv_file(table_or_database: Table | Database, file: str):
    """Export database/table to csv file, will act different if database is exported.

    If database is passed, it tries to create a directory then export all table"""
    readed = to_csv_string(table_or_database)
    if isinstance(readed, list):
        if not exists(file):
            mkdir(file)
        if isfile(file):
            raise NotADirectoryError(f"{file} is a file, export failed!")

        _process_database(file, readed)
        return True
    with open(file, "w", encoding="utf-8") as fio:
        fio.write(readed)
    return True

def from_csv_string(table: Table, data: str):
    """Insert from CSV data"""
    reader = DictReader(data)
    for entry in reader:
        table.insert(entry)

def from_csv_file(table: Table, data: str):
    """Insert from CSV file"""
    with open(data, encoding='utf-8') as f:
        fdata = f.read()
    return from_csv_string(table, fdata)



"""Query Builder"""

from .core import build_delete, build_insert, build_select, build_update
from .engine import build_update_data, SubQuery

__all__ = [
    "build_delete",
    "build_insert",
    "build_select",
    "build_update",
    "build_update_data",
]



"""Table API update test"""

from pytest import raises
from luminadb import Database
from luminadb.operators import op
from luminadb.errors import TableRemovedError

from ..setup import (
    database,
    groups,
    users,
    setup_database_builder,
    save_report,
    setup_database_fns,
    GROUP_UNEW,
    USER_UNEW,
)


def test_update():
    """test 0201 update"""
    db = Database(":memory:")
    setup_database_builder(db)
    usrs = db.table("users")
    grps = db.table("groups")
    assert grps.update({"id": op == 1}, GROUP_UNEW[0]) == 1
    assert usrs.update({"id": op == 1}, USER_UNEW[0]) == 1
    assert save_report("02_update", db, grps, usrs)


def test_update_one():
    """test 0202 update one"""
    db = Database(":memory:")
    setup_database_builder(db)
    usrs = db.table("users")
    grps = db.table("groups")
    assert grps.update_one({"id": op == 1}, GROUP_UNEW[0]) == 1
    assert usrs.update_one({"id": op == 1}, USER_UNEW[0]) == 1
    assert save_report("02_update", db, grps, usrs)


def test_update_limited_order():
    """Test 0203 update limited with order"""
    db = Database(":memory:")
    setup_database_fns(db)
    checkout = db.table("checkout")
    assert (
        checkout.update({"quantity": 50}, {"quantity": 70}, limit=2) == 2
    )  # type: ignore
    assert len(checkout.select({"quantity": 70}, limit=2)) == 2

def test_finish():
    """test 0300 finish"""
    with raises(TableRemovedError):
        assert save_report("04_finish", database, groups, users)
        assert database.delete_table("groups") is None
        assert database.delete_table("users") is None
        assert save_report("05_finish", database, groups, users)



"""Subexpression"""

from __future__ import annotations
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from .table import Table
    from .query_builder.typings import Condition, OnlyColumn
    from .typings import Orders


class SubQuery:
    """Allows subquery usage. Only select at this time"""

    def __init__(
        self,
        table: str | Table,
        cols: OnlyColumn,
        where: Condition,
        limit: int = 0,
        order: Orders | None = None,
    ) -> None:
        self._table = table if isinstance(table, str) else table.name
        self._cols = cols
        self._where = (
            tuple(where.items() if isinstance(where, dict) else where)
            if where is not None
            else ()
        )
        self._limit = limit
        self._order = order

    def __hash__(self):
        return hash((self._table, self._cols, self._where, self._limit))

    def get_params(self):
        """Get parameters/where data"""
        return tuple((value for key, value in self._where))

    @property
    def table(self):
        """Return target table name"""
        return self._table

    @property
    def cols(self):
        """Return columns associated"""
        return self._cols

    @property
    def where(self):
        """Return all conditions"""
        return self._where

    @property
    def limit(self):
        """Return limit query"""
        return self._limit

    @property
    def orders(self):
        """Return orders"""
        return self._order



"""SQLite Database"""

from atexit import register as finalize
from sqlite3 import OperationalError, connect
from typing import Iterable, Literal, Optional

from luminadb._debug import if_debug_print

from .utils import (
    WithCursor,
    check_iter,
    check_one,
    dict_factory,
    sqlite_multithread_check,
)
from .column import BuilderColumn, Column
from .query_builder.table_creation import extract_table_creations
from .table import Table
from .errors import DatabaseExistsError, DatabaseMissingError
from .index import Index

Columns = Iterable[Column] | Iterable[BuilderColumn]

__all__ = ["Database"]

IGNORE_TABLE_CHECKS = ("sqlite_master", "sqlite_temp_schema", "sqlite_temp_master")


class Database: # pylint: disable=too-many-instance-attributes
    """Sqlite3 database, this provide basic integration.

    Custom flags:
        strict : Certain actions are prevented when active, i.e, initializing nonexistent tables
        forgive: Certain actions are replaced when active, i.e, replacing .create_table to .table
                 when a table exists"""

    def __init__(self, path: str, **kwargs) -> None:
        kwargs["check_same_thread"] = sqlite_multithread_check() != 3
        self._path = path
        self._strict: bool = kwargs.get("strict", True)
        self._forgive: bool = kwargs.get("forgive", True)
        self._active = []
        if 'forgive' in kwargs:
            del kwargs['forgive']
        if 'strict' in kwargs:
            del kwargs['strict']
        self._config = None
        self._closed = False
        self._table_class = Table
        if not self._closed or self.__dict__.get("_initiated", False) is False:
            finalize(self._finalizer)
            self._initiated = True
        self._kwargs = kwargs
        self._create_connection()

    def _create_connection(self):
        self._database = connect(self._path, **self._kwargs)
        self._database.row_factory = dict_factory

    def _finalizer(self):
        self.close()

    def cursor(self) -> WithCursor:
        """Create cursor"""
        return self._database.cursor(WithCursor)  # type: ignore

    def create_table(self, table: str, columns: Columns):
        """Create table

        Args:
            table (str): Table name
            columns (Iterable[Column]): Columns for table

        Returns:
            Table: Newly created table
        """
        columns = (
            column.to_column() if isinstance(column, BuilderColumn) else column
            for column in columns
        )
        tbquery = extract_table_creations(columns)
        query = f"create table {table} ({tbquery})"

        if_debug_print(query)

        if self._forgive and self.check_table(table):
            return self.table(table, columns)

        try:
            cursor = self._database.cursor()
            cursor.execute(query)
            self._database.commit()
        except OperationalError as error:
            if "already exists" in str(error):
                dberror = DatabaseExistsError(f"table {table} already exists.")
                dberror.add_note(f"{type(error).__name__}: {error!s}")
                raise dberror from error
            error.add_note(f"Query: {query}")
            raise error
        table_ = self.table(table, columns)
        table_._deleted = False  # pylint: disable=protected-access
        return table_

    def delete_table(self, table: str):
        """Delete an existing table

        Args:
            table (str): table name
        """
        check_one(table)
        table_ = self.table(table)
        self._database.cursor().execute(f"drop table {table}")
        # pylint: disable-next=protected-access
        table_._delete_hook()  # pylint: disable=protected-access

    def table(self, table: str, __columns: Optional[Iterable[Column]] = None):  # type: ignore
        """fetch table"""

        if self._strict and not self.check_table(table):
            raise DatabaseMissingError(f"table {table} does not exists.")

        try:
            this_table = self._table_class(self, table, __columns)
        except OperationalError as exc:
            dberror = DatabaseMissingError(f"table {table} does not exists")
            dberror.add_note(f"{type(exc).__name__}: {exc!s}")
            raise dberror from None
        return this_table

    def reset_table(self, table: str, columns: Columns) -> Table:
        """Reset existing table with new, this rewrote entire table than altering it."""
        try:
            self.delete_table(table)
        except OperationalError:
            pass
        return self.create_table(table, columns)

    def rename_table(self, old_table: str, new_table: str) -> Table:
        """Rename existing table to a new one."""
        check_iter((old_table, new_table))
        cursor = self.sql.cursor()
        cursor.execute(f"alter table {old_table} rename to {new_table}")
        self.sql.commit()
        return self.table(new_table)

    def check_table(self, table: str):
        """Check if table is exists or not."""
        # if self._path in PLUGINS_PATH:
        #     plugin = self._path[2:]
        #     raise ValueError(f"Plugin {plugin} must redefine check_table.")
        check_one(table)
        if table in IGNORE_TABLE_CHECKS:
            return True  # Let's return true.
        cursor = self.sql.cursor()
        cursor.execute(
            "select name from sqlite_master where type='table' and name=?", (table,)
        )
        return cursor.fetchone() is not None

    def create_index(self, index: Index):
        """Create an index"""
        with self.sql as dbcursor:
            dbcursor.execute(index.build_sql())

    def delete_index(self, name: str | Index, exists_ok: bool = False):
        """Drop an index"""
        _name = name.index_name if isinstance(name, Index) else name
        check_one(_name)
        with self.sql as dbcursor:
            if_ok = "if exists" if exists_ok else ""
            dbcursor.execute(f"drop index {if_ok} {_name}")

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {id(self)}>"

    def close(self):
        """Close database"""
        if self._closed:
            return
        self._database.close()
        if self.path == ":memory:":
            self._closed = True
            return
        self._closed = True

    def tables(self) -> tuple[Table, ...]:
        """Return tuple containing all table except internal tables"""
        master = self.table("sqlite_master")
        listed = []
        for table in master.select():
            if table.type == "table":
                listed.append(self.table(table.name))
        return tuple(listed)

    def commit(self):
        """Commit changes to database"""
        self._database.commit()

    def rollback(self):
        """Rollback changes"""
        self._database.rollback()

    def foreign_pragma(self, bool_state: Literal["ON", "OFF", ""] = ""):
        """Enable/disable foreign key pragma"""
        if bool_state not in ("ON", "OFF", ""):
            raise ValueError("Either ON/OFF for foreign key pragma.")
        return self._database.execute(
            f"PRAGMA foreign_keys{'='+bool_state if bool_state else ''}"
        ).fetchone()  # pylint: disable=line-too-long

    def optimize(self):
        """Optimize current database"""
        return self._database.execute("PRAGMA optimize").fetchone()

    def shrink_memory(self):
        """Shrink memories from database as much as it can."""
        return self._database.execute("PRAGMA shrink_memory").fetchone()

    def vacuum(self):
        """Vacuum this database"""
        return self._database.execute("VACUUM").fetchone()

    @property
    def closed(self):
        """Is database closed?"""
        return self._closed

    @closed.setter
    def closed(self, __o: bool):
        """Is database closed?"""
        if __o:
            self.close()
            return
        raise ValueError("Expected non-false/non-null value")

    @property
    def path(self):
        """Path to SQL Connection"""
        return self._path or ":memory:"

    @property
    def sql(self):
        """SQL Connection"""
        return self._database



"""Core database utility"""
# pylint: disable=useless-parent-delegation

# This module must not import anything from this package except errors.

from uuid import uuid4
from re import Pattern
from re import compile as re_compile
from re import escape as re_escape
from sqlite3 import Cursor, connect
from string import punctuation
from typing import Any, Generic, Iterable, Iterator, Mapping, TypeVar, TypeAlias

from .errors import SecurityError


T = TypeVar("T")
_INVALID_STR = punctuation.replace("_", "")
_re_valid = re_compile(f"[{re_escape(_INVALID_STR)}]+")

_SQLITE_KEYWORDS = {
    "ABORT",
    "ACTION",
    "ADD",
    "AFTER",
    "ALL",
    "ALTER",
    "ANALYZE",
    "AND",
    "AS",
    "ASC",
    "ATTACH",
    "AUTOINCREMENT",
    "BEFORE",
    "BEGIN",
    "BETWEEN",
    "BY",
    "CASCADE",
    "CASE",
    "CAST",
    "CHECK",
    "COLLATE",
    "COLUMN",
    "COMMIT",
    "CONFLICT",
    "CONSTRAINT",
    "CREATE",
    "CROSS",
    "CURRENT_DATE",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "DATABASE",
    "DEFAULT",
    "DEFERRABLE",
    "DEFERRED",
    "DELETE",
    "DESC",
    "DETACH",
    "DISTINCT",
    "DROP",
    "EACH",
    "ELSE",
    "END",
    "ESCAPE",
    "EXCEPT",
    "EXCLUSIVE",
    "EXISTS",
    "EXPLAIN",
    "FAIL",
    "FOR",
    "FOREIGN",
    "FROM",
    "FULL",
    "GLOB",
    "GROUP",
    "HAVING",
    "IF",
    "IGNORE",
    "IMMEDIATE",
    "IN",
    "INDEX",
    "INDEXED",
    "INITIALLY",
    "INNER",
    "INSERT",
    "INSTEAD",
    "INTERSECT",
    "INTO",
    "IS",
    "ISNULL",
    "JOIN",
    "KEY",
    "LEFT",
    "LIKE",
    "LIMIT",
    "MATCH",
    "NATURAL",
    "NO",
    "NOT",
    "NOTNULL",
    "NULL",
    "OF",
    "OFFSET",
    "ON",
    "OR",
    "ORDER",
    "OUTER",
    "PLAN",
    "PRAGMA",
    "PRIMARY",
    "QUERY",
    "RAISE",
    "RECURSIVE",
    "REFERENCES",
    "REGEXP",
    "REINDEX",
    "RELEASE",
    "RENAME",
    "REPLACE",
    "RESTRICT",
    "RIGHT",
    "ROLLBACK",
    "ROW",
    "SAVEPOINT",
    "SELECT",
    "SET",
    "TABLE",
    "TEMP",
    "TEMPORARY",
    "THEN",
    "TO",
    "TRANSACTION",
    "TRIGGER",
    "UNION",
    "UNIQUE",
    "UPDATE",
    "USING",
    "VACUUM",
    "VALUES",
    "VIEW",
    "VIRTUAL",
    "WHEN",
    "WHERE",
    "WITH",
    "WITHOUT",
}


class NullObject:
    """Null object"""

    __obj = None

    def __new__(cls):
        if cls.__obj is None:
            cls.__obj = object.__new__(cls)
        return cls.__obj

    def __init__(self) -> None:
        pass

    def __repr__(self) -> str:
        return "null"

    def __str__(self) -> str:
        return repr(self)

    def __int__(self) -> int:
        return 0

    def __float__(self) -> float:
        return 0.0


class Sentinel:  # pylint: disable=too-few-public-methods
    """A Sentinel value, has unique behavior for Table API and Model API.

    Pre-defined value:
        - SKIP, use this one if you're unsure if the data you pulled exists or not.
            The query builder will remove it if it detects SKIP sentinel."""

    def __repr__(self) -> str:
        return "<Sentinel>"


class Row(Mapping[str, T], Generic[T], dict): # type: ignore
    """Attribute Dictionary"""

    def __setitem__(self, key: str, value: T):
        super().__setitem__(key, value)

    def __getitem__(self, key: str) -> T:
        return dict.__getitem__(self, key)

    def __iter__(self) -> Iterator[str]:
        return dict.__iter__(self)

    def __len__(self) -> int:
        return dict.__len__(self)

    def __getattr__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            return self[__name]


def dict_factory(cursor, row):
    """dict factory"""
    fields = [column[0] for column in cursor.description]
    return Row(zip(fields, row))
    # return {key: value for key, value in zip(fields, row)}


def sqlite_multithread_check():
    """sqlite mulththread check"""
    thread_safe = {0: 0, 2: 1, 1: 3}
    conn = connect(":memory:", check_same_thread=False)
    data = conn.execute(
        "select * from pragma_compile_options where compile_options \
like 'THREADSAFE=%'"
    ).fetchone()[0]
    conn.close()

    threadsafety_value = int(data.split("=")[1])
    return thread_safe[threadsafety_value]


def matches(pattern: Pattern, value: str):
    """matches"""
    if len(pattern.findall(value)) == 0:
        return False
    return True


def check_one(data: str, bypass_list: tuple[str, ...] | None = None):
    """check one to check if a string contains illegal character OR
    if it is a reserved SQL keyword"""
    if matches(_re_valid, data) is True:
        exc = SecurityError("Cannot parse unsafe data.")
        exc.add_note(f"Target: {data}")
        raise exc
    bp = bypass_list or []
    if data.upper() in _SQLITE_KEYWORDS:
        if data in bp:
            return data
        raise SecurityError(f'"{data}" is a reserved SQL keyword and cannot be used.')
    return data


def check_iter(data: Iterable[str], bypass_list: tuple[str, ...] | None = None):
    """An iterable checks as it's check_one"""
    for val in data:
        check_one(val, bypass_list)


class WithCursor(Cursor):
    """With cursor"""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type and exc_value and traceback:
            return True
        self.connection.commit()
        return True

    def __repr__(self) -> str:
        return type(self).__name__


class NoopResource:
    """No-op resource control"""

    def close(self):
        """Close this resource"""

    def open(self, *_, **__):
        """open something"""


def test_installed():
    """Test if project is truly installed. It was meant for docs compatibility and thus,
    always returns true"""
    return True

def generate_ids():
    """Generate ids for statements"""
    return str(uuid4().int)


null = NullObject()
SKIP = Sentinel()
AttrDict = Row
Query: TypeAlias = Row[Any]
SquashedQueries: TypeAlias = Row[list[Any]]
Queries: TypeAlias = list[Query] | SquashedQueries

def crunch(query: Queries | list[dict[str, Any]]) -> Row[list[Any]]:
    """Crunch queries into Row that all of its value become a list."""
    data: Row[list[Any]] = Row()
    for value in query:
        for key, val in value.items(): # type: ignore
            if key not in data:
                data[key] = []
            data[key].append(val)
    return data



__all__ = [
    "null",
    "SKIP",
    "WithCursor",
    "check_iter",
    "check_one",
    "matches",
    "Row",
    "AttrDict",
    "NullObject",
    "sqlite_multithread_check",
    "NoopResource",
]



"""Column"""
# pylint: disable=redefined-builtin

from json import dumps, loads
from typing import Any, Self, Callable
from .locals import _PATH, SQLACTION
from .utils import check_one, matches
from .typings import null


class Column:  # pylint: disable=too-many-instance-attributes
    """Column

    tip: for foreign_ref, you can split with / to separate table and column name.
    e.g: user/id"""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: str,
        type: str,
        foreign: bool = False,
        foreign_ref: str | None = None,
        primary: bool = False,
        unique: bool = False,
        nullable: bool = True,
        default: Any = None,
        on_delete: SQLACTION = "cascade",
        on_update: SQLACTION = "cascade",
    ) -> None:
        self._name = check_one(name)
        self._type = check_one(type)
        self._unique = unique
        self._nullable = nullable
        self._default = default
        self._foreign_enabled = foreign
        while foreign_ref:
            if not matches(_PATH, foreign_ref):
                raise ValueError(
                    "foreign_ref has no / separator to separate table and column."
                )
            ref = foreign_ref.split("/", 1)
            source = ref[0]
            scolumn = ref[1] if len(ref) == 2 else name
            self._source = source
            self._source_column = scolumn
            self._foreign = foreign_ref
            break

        if not foreign:
            self._foreign = None
            self._foreign_enabled = False
        self._update = on_update
        self._delete = on_delete
        self._is_primary = primary

    @property
    def name(self):
        """Column Name"""
        return self._name

    @property
    def unique(self):
        """Is unique"""
        return self._unique

    @property
    def default(self):
        """Default value"""
        return self._default

    @property
    def nullable(self):
        """Nullable"""
        return self._nullable

    @property
    def raw_source(self):
        """Source / Foreign Reference"""
        return self._foreign

    @property
    def foreign(self):
        """Is foreign enabled?"""
        return self._foreign_enabled

    @property
    def source(self):
        """Source / Foreign Reference"""
        if self._foreign is None:
            raise AttributeError("Source is unset")
        return self._source

    @property
    def source_column(self):
        """Source column / Foreign reference column"""
        if self._foreign is None:
            raise AttributeError("Source column is unset")
        return self._source_column

    @property
    def primary(self):
        """Is primary or not?"""
        return self._is_primary

    @property
    def on_update(self):
        """Update setting"""
        return self._update

    @property
    def on_delete(self):
        """Delete setting"""
        return self._delete

    @property
    def type(self):
        """Type"""
        return self._type

    def __repr__(self) -> str:
        return f"<{self.type.title()}{type(self).__name__} -> {self.name}>"

    def __eq__(self, __o: "Column") -> bool:
        if not isinstance(__o, Column):
            raise NotImplementedError
        other = (
            __o.name,
            __o.type,
            __o.unique,
            __o.nullable,
            __o.default,
            __o.primary,
            __o.raw_source,
            __o.on_delete,
            __o.on_update,
        )
        self_ = (
            self.name,
            self.type,
            self.unique,
            self.nullable,
            self.default,
            self.primary,
            self.raw_source,
            self.on_delete,
            self.on_update,
        )
        return all((item1 in self_ for item1 in other))

    def to_json(self):
        """Convert column info as JSON"""
        return dumps({
            "name": self.name,
            "type": self.type,
            "iunique": self.unique,
            "primary": self.primary,
            "foreign": self.foreign,
            "foreign_ref": self.raw_source,
            "nullable": self.nullable,
            "default": self.default,
            "on_delete": self.on_delete,
            "on_update": self.on_update
        })

    @classmethod
    def from_json(cls, json: str):
        """Convert JSON to a column"""
        return cls(**loads(json))

class BuilderColumn:  # pylint: disable=too-many-instance-attributes
    """Builder Column -- Column Implementation using Builder Column"""

    def __init__(self, _from_typelist: list[str] | None = None) -> None:
        self._update: SQLACTION = "cascade"
        self._delete: SQLACTION = "cascade"
        self._primary = False
        self._default = None
        self._nullable = False
        self._unique = False
        self._type: str | None = None  # type: ignore
        self._name = ""
        self._source = ""
        self._source_column = ""
        self._foreign = False
        self._types: list[str] = _from_typelist or []
        self._has_defined_type = False

    def _check_then_define(self):
        if self._has_defined_type:
            error = ValueError(f"Cannot reset current type of {self._type}")
            error.add_note("Consider removing calls that refer as types.")
            raise error
        self._has_defined_type = True

    def _set_name(self, name: str) -> Self:
        self._name = name
        return self

    def set_type(self, name: str) -> Callable[[str], Self]:
        """Set type from name, this will be handled by class' type checking.
        This is a setup function. You need to do .set_type(type_name)(name) to do
        similiar effect to .integer(name) as for example"""
        self._check_then_define()
        if not self._types:
            self._type = name
            return self._set_name
        if name not in self._types:
            self._has_defined_type = False
            self._type = None
            raise TypeError(f"{name} was not defined.")
        self._type = name
        return self._set_name

    def default(self, default_value: Any):
        """Set default value"""
        if not default_value:
            self._nullable = True
        self._default = default_value
        return self

    def primary(self) -> Self:
        """Set primary"""
        self._primary = True
        self._unique = True
        return self

    def unique(self) -> Self:
        """Set unique"""
        self._unique = True
        return self

    def foreign(self, source: str) -> Self:
        """Set foreign reference"""
        if not "/" in source:
            raise ValueError("Foreign ref invalid")
        self._source, self._source_column = source.split("/", 1)
        self._foreign = True
        return self

    def on_update(self, action: SQLACTION):
        """Set on update action"""
        self._update = action
        return self

    def allow_null(self):
        """Allow null"""
        self._nullable = True
        return self

    def on_delete(self, action: SQLACTION):
        """Set on delete action"""
        self._delete = action
        return self

    def _column_challenge(self):
        if not self._name:
            raise ValueError("Name must be defined")
        if not self._type:
            raise ValueError("Type must be defined")
        if self._foreign:
            if not self._source or self._source_column:
                raise ValueError("One of foreign ref must present")
        if not self._nullable and self._default is null:
            raise ValueError("Cannot set from not null while default is null")

    def to_column(self):
        """Conver from BuilderColumn to Column"""
        if not self._type:
            raise ValueError("type must be defined first")
        return Column(
            self._name,
            self._type,
            self._foreign,
            self._source + "/" + self._source_column,
            self._primary,
            self._unique,
            self._nullable,
            self._default,
            self._delete,
            self._update,
        )

    def __eq__(self, __o: "Column") -> bool:  # type: ignore
        return self.to_column() == __o

    def __repr__(self) -> str:
        return f"<{self._type} of {self._name}>"


def text(name: str) -> BuilderColumn:
    """Create a text column with name"""
    return BuilderColumn().set_type("text")(name)


def integer(name: str) -> BuilderColumn:
    """Create a integer column with name"""
    return BuilderColumn().set_type("integer")(name)


def blob(name: str) -> BuilderColumn:
    """Create a blob column with name"""
    return BuilderColumn().set_type("blob")(name)


def real(name: str) -> BuilderColumn:
    """Create a real column with name"""
    return BuilderColumn().set_type("real")(name)

def boolean(name: str) -> BuilderColumn:
    """Create a boolean column with name"""
    return BuilderColumn().set_type("boolean")(name)


def create_calls(typename: str, types: list[str]) -> Callable[[str], BuilderColumn]:
    """Create a dynamic call for types. This is intended to mimic
    this module's real(), text(), ... with custom types."""
    return BuilderColumn(types).set_type(typename)


__all__ = ["Column", "BuilderColumn", "text", "integer", "blob", "real", "create_calls"]



"""Mixin Helpers"""

# pylint: disable=too-few-public-methods

from typing import TypeVar, Callable, Type, Generator
from . import BaseModel

BaseModelT = TypeVar("BaseModelT", bound=BaseModel)


class BaseModelMixin:
    """Base class for all Mixins"""


class ScopeMixin(BaseModelMixin):
    """Scope-related mixins"""

    @classmethod  # type: ignore
    def active(cls: Type[BaseModelT]) -> list[BaseModelT]:  # type: ignore
        """Return any active users"""
        return cls.where(is_active=True).fetch()


class ChunkableMixin(BaseModelMixin):
    """Implement chunk related stuff"""

    @classmethod
    def chunk_callback(
        cls: Type[BaseModelT],  # type: ignore
        __limit: int,
        __callback: Callable[[list[BaseModelT]], None],
        /,
        **kwargs,
    ):
        """Return specified instance by the amount of limit, or execute provided callback"""
        offset = 0
        while True:
            batch = cls.where(**kwargs).limit(__limit).offset(offset).fetch()

            if not batch:
                break

            __callback(batch)

            if len(batch) != __limit:
                break

            offset += __limit

    @classmethod  # type: ignore
    def chunk(
        cls: Type[BaseModelT], __limit: int, /, **kwargs  # type: ignore
    ) -> Generator[list[BaseModelT], None, None]:  # type: ignore
        """Yield specifiec range instanced by the amount of limit."""
        offset = 0
        while True:
            batch = cls.where(**kwargs).limit(__limit).offset(offset).fetch()

            if not batch:
                break

            yield batch

            if len(batch) != __limit:
                break

            offset += __limit



"""Table"""

# pylint: disable=too-many-arguments,too-many-public-methods,R0801

from contextvars import ContextVar
from sqlite3 import Connection, Error, OperationalError
from typing import (
    Any,
    Generator,
    Iterable,
    Literal,
    Optional,
    overload,
    TYPE_CHECKING
)

from luminadb.functions import ParsedFn, Function, count
from luminadb.subquery import SubQuery


from .utils import check_iter, check_one, Row, crunch
from ._debug import if_debug_print
from .column import BuilderColumn, Column
from .errors import TableRemovedError
from .query_builder import (
    # extract_single_column,
    # fetch_columns,
    build_select,
    build_insert,
    build_delete,
    build_update,
)
from .query_builder.typings import Condition
from .query_builder.table_creation import extract_single_column
# from .signature import op
from .typings import (
    Data,
    Orders,
    Query,
    # _MasterQuery,
    OnlyColumn,
    SquashedQueries,
    JustAColumn,
)

if TYPE_CHECKING:
    from .database import Database

# Let's add a little bit of 'black' magic here.
_null = Function("__NULL__")()
_tx_stack = ContextVar("_tx_stack", default=[])

class Table: # pylint: disable=too-many-instance-attributes
    """Table. Make sure you remember how the table goes."""


    def __init__(
        self,
        parent: "Database",  # type: ignore
        table: str,
        columns: Optional[Iterable[Column]] = None,  # type: ignore
    ) -> None:
        if parent.closed:
            raise ConnectionError("Connection to database is already closed.")
        self._parent_repr = repr(parent)
        self._db = parent
        self._sql: Connection = parent.sql
        # pylint: disable-next=protected-access
        self._sql_path = parent._path
        self._deleted = False
        self._force_dirty = False
        self._dirty = False
        self._auto = True
        self._table = check_one(table)
        self._prev_autocommit = None
        self._prev_auto = True
        self._columns: Optional[list[Column]] = list(columns) if columns else None

    def __enter__(self):
        self._prev_auto = self._auto
        self._prev_autocommit = self._sql.isolation_level

        self._auto = False
        self._sql.isolation_level = None
        self._begin_transaction()
        return self

    def __exit__(self, exc_type, _, __):
        if exc_type is None:
            self._commit_transaction()
            self._dirty = False
        else:
            self._rollback_transaction()
            self._dirty = False

        self._sql.isolation_level = self._prev_autocommit
        self._auto = self._prev_auto

    @property
    def deleted(self):
        """Is table deleted"""
        return self._deleted

    @property
    def name(self):
        """Table name"""
        return self._table

    @property
    def force_dirty(self):
        """Force dirty state, whether .selecting() on dirty/uncommitted data is allowed or not"""
        return self._force_dirty

    @force_dirty.setter
    def force_dirty(self, value: bool):
        """Force dirty state, whether .selecting() on dirty/uncommitted data is allowed or not"""
        if not isinstance(value, bool):
            return
        self._force_dirty = value

    @property
    def auto_commit(self):
        """Auto commit state of this instance"""
        return self._auto

    @auto_commit.setter
    def auto_commit(self, value: bool):
        if not isinstance(value, bool):
            return
        self._auto = value

    @property
    def in_transaction(self):
        """Returns True if the table is in an active transaction."""
        return not self._auto or self._sql.isolation_level is None

    def _finalize(self):
        pass

    def _delete_hook(self):
        try:
            self.select()
        except OperationalError:
            self._deleted = True

    def _exec(
        self,
        query: str,
        data: dict[str, Any] | list[dict[str, Any]],
        which: Literal["execute", "executemany"] = "execute",
    ):
        """Execute a sql query"""
        if_debug_print(query, '\n', data)
        cursor = self._sql.cursor()
        fn = cursor.execute if which == "execute" else cursor.executemany
        try:
            fn(query, data)
        except Error as exc:
            if str(exc).startswith("no such table:"):
                raise TableRemovedError(f"Table {self._table} doesn't exists anymore") from None
            exc.add_note(f"SQL query: {query}")
            exc.add_note(f"Arguments: {data}")
            exc.add_note(
                f"There's about {1 if isinstance(data, dict) else len(data)} value(s) inserted"
            )
            raise exc
        return cursor

    def _control(self):
        if self._deleted:
            raise TableRemovedError(f"{self._table} is already removed")

    def _query_control(self):
        if self._dirty and self._force_dirty is False:
            self._sql.commit()
            self._dirty = False

    def force_nodelete(self):
        """Force "undelete" table. Used if table was mistakenly assigned as
        deleted."""
        self._deleted = True

    def delete(
        self,
        where: Condition = None,
        limit: int = 0,
        order: Optional[Orders] = None,
    ):
        """Delete row or rows

        Args:
            where (Condition, optional): Condition to determine deletion
                See `Signature` class about conditional stuff. Defaults to None.
            limit (int, optional): Limit deletion by integer. Defaults to 0.
            order (Optional[Orders], optional): Order of deletion. Defaults to None.

        Returns:
            int: Rows affected
        """
        query, data = build_delete(self._table, where, limit, order)  # type: ignore
        self._control()
        cursor = self._exec(query, data)
        rcount = cursor.rowcount
        if not self.in_transaction:
            self._sql.commit()
        else:
            self._dirty = True
        return rcount

    def delete_one(self, where: Condition = None, order: Optional[Orders] = None):
        """Delete a row

        Args:
            where (Condition, optional): Conditional to determine deletion.
            Defaults to None.
            order (Optional[Orders], optional): Order of deletion. Defaults to None.
        """
        return self.delete(where, 1, order)

    def insert(self, data: Data):
        """Insert data to current table

        Args:
            data (Data): Data to insert. Make sure it's compatible with the table.

        Returns:
            int: Last rowid
        """
        query, _ = build_insert(self._table, data)  # type: ignore
        self._control()
        cursor = self._exec(query, data)
        rlastrowid = cursor.lastrowid
        if not self.in_transaction:
            self._sql.commit()
        else:
            self._dirty = True
        return rlastrowid

    def insert_multiple(self, datas: list[Data]):
        """Insert multiple values

        Args:
            datas (Iterable[Data]): Data to be inserted.
        """
        self._control()
        query, _ = build_insert(self._table, datas[0])  # type: ignore
        self._exec(query, datas, "executemany")
        if not self.in_transaction:
            self._sql.commit()
        else:
            self._dirty = True

    def insert_many(self, datas: list[Data]):
        """Alias to `insert_multiple`"""
        return self.insert_multiple(datas)

    def update(
        self,
        where: Condition | None = None,
        data: Data | None = None,
        limit: int = 0,
        order: Optional[Orders] = None,
    ):
        """Update rows of current table

        Args:
            data (Data): New data to update
            where (Condition, optional): Condition dictionary.
                See `Signature` about how condition works. Defaults to None.
            limit (int, optional): Limit updates. Defaults to 0.
            order (Optional[Orders], optional): Order of change. Defaults to None.

        Returns:
            int: Rows affected
        """
        if data is None:
            raise ValueError("data parameter must not be None")
        query, data = build_update(
            self._table, data, where, limit, order
        )  # type: ignore
        self._control()
        cursor = self._exec(query, data)
        rcount = cursor.rowcount
        if not self.in_transaction:
            self._sql.commit()
        else:
            self._dirty = True
        return rcount

    def update_one(
        self,
        where: Condition | None = None,
        data: Data | None = None,
        order: Orders | None = None,
    ) -> int:
        """Update 1 data only"""
        return self.update(where, data, 1, order)

    @overload
    def select(
        self,
        where: Condition = None,
        what: OnlyColumn = "*",
        limit: int = 0,
        offset: int = 0,
        order: Optional[Orders] = None,
        flatten: Literal[False] = False,
    ) -> list[Query]:
        pass

    @overload
    def select(
        self,
        where: Condition = None,
        what: OnlyColumn = "*",
        limit: int = 0,
        offset: int = 0,
        order: Optional[Orders] = None,
        flatten: Literal[True] = True,
    ) -> SquashedQueries:
        pass

    @overload
    def select(
        self,
        where: Condition = None,
        what: ParsedFn = _null,
        limit: int = 0,
        offset: int = 0,
        order: Optional[Orders] = None,
        flatten: Literal[False] = False,
    ) -> Any:
        pass

    @overload
    def select(
        self,
        where: Condition = None,
        what: JustAColumn = "_COLUMN",
        limit: int = 0,
        offset: int = 0,
        order: Optional[Orders] = None,
        flatten: Literal[False] = False,
    ) -> list[Any]:
        pass

    def select(
        self,  # pylint: disable=too-many-arguments
        where: Condition = None,
        what: OnlyColumn | ParsedFn | JustAColumn = "*",
        limit: int = 0,
        offset: int = 0,
        order: Optional[Orders] = None,
        flatten: bool = False,
    ):
        """Select data in current table. Bare .select() returns all data.

        Args:
            where (Condition, optional): Conditions to used. Defaults to None.
            what: (OnlyColumn, ParsedFn, optional): Select what you want. Default to None.
            limit (int, optional): Limit of select. Defaults to 0.
            offset (int, optional): Offset. Defaults to 0
            order (Optional[Orders], optional): Selection order. Defaults to None.
            flatten (bool): Flatten returned data into dict of lists. Defaults to False.

        Returns:
            Queries: Selected data
        """
        self._control()
        self._query_control()
        query, data = build_select(
            self._table, where, what, limit, offset, order
        )  # type: ignore
        just_a_column = (isinstance(what, tuple) and len(what) == 1) or (
            isinstance(what, str) and what != "*"
        )
        with self._sql:
            cursor = self._exec(query, data)
            data = cursor.fetchall()
            if just_a_column:
                return [d[what] for d in data]
            if flatten:
                return crunch(data)
            if isinstance(what, ParsedFn):
                return data[0][what.parse_sql()[0]]
            return data

    @overload
    def paginate_select(
        self,
        where: Condition = None,
        what: OnlyColumn = "*",
        page: int = 0,
        length: int = 10,
        order: Optional[Orders] = None,
        flatten: Literal[False] = False,
    ) -> Generator[list[Query], None, None]:
        pass

    @overload
    def paginate_select(
        self,
        where: Condition = None,
        what: JustAColumn = "_COLUMN",
        page: int = 0,
        length: int = 10,
        order: Optional[Orders] = None,
        flatten: Literal[False] = False,
    ) -> Generator[list[Any], None, None]:  # type: ignore
        pass

    @overload
    def paginate_select(
        self,
        where: Condition = None,
        what: OnlyColumn = "*",
        page: int = 0,
        length: int = 10,
        order: Optional[Orders] = None,
        flatten: Literal[True] = True,
    ) -> Generator[SquashedQueries, None, None]:
        pass

    def paginate_select(
        self,
        where: Condition = None,
        what: OnlyColumn | JustAColumn = "*",
        page: int = 0,
        length: int = 10,
        order: Optional[Orders] = None,
        flatten: bool = False,
    ):
        """Paginate select

        Args:
            where (Condition, optional): Confitions to use. Defaults to None.
            what (OnlyColumn, optional): Select what you want. Default to None.
            page (int): Which page number be returned first
            length (int, optional): Pagination length. Defaults to 10.
            order (Optional[Orders], optional): Order. Defaults to None.
            flatten (bool): Flatten returned data into dict of lists. Defaults to False.

        Yields:
            Generator[Queries, None, None]: Step-by-step paginated result.
        """

        if page < 0:
            page = 0
            order = "desc" if order in ("asc", None) else "asc"  # type: ignore
        self._control()
        self._query_control()
        start = page * length
        # ! A `only` keyword as a string or tuple of 1 element will
        # ! actually be a problem if they left alone because the end result is a list
        just_a_column = (isinstance(what, str) and what != "*") or (
            isinstance(what, tuple) and len(what) == 1
        )
        while True:
            query, data = build_select(
                self._table, where, what, length, start, order
            )  # type: ignore
            with self._sql:
                cursor = self._exec(query, data)
                fetched = cursor.fetchmany(length)
                if len(fetched) == 0:
                    return
                if flatten and not just_a_column:
                    fetched = crunch(fetched)
                if len(fetched) != length:
                    yield fetched
                    return
                yield fetched
                start += length

    @overload
    def select_one(
        self,
        where: Condition = None,
        what: ParsedFn = _null,
        order: Optional[Orders] = None,
    ) -> Any:
        pass

    @overload
    def select_one(
        self,
        where: Condition = None,
        what: OnlyColumn = "*",
        order: Optional[Orders] = None,
    ) -> Query:
        pass

    @overload
    def select_one(
        self,
        where: Condition = None,
        what: JustAColumn = "_COLUMN",
        order: Optional[Orders] = None,
    ) -> Any:
        pass

    def select_one(
        self,
        where: Condition = None,
        what: OnlyColumn | JustAColumn | ParsedFn = "*",
        order: Optional[Orders] = None,
    ):
        """Select one data

        Args:
            where (Condition, optional): Condition to use. Defaults to None.
            what: (OnlyColumn, optional): Select what you want. Default to None.
            order (Optional[Orders], optional): Order of selection. Defaults to None.

        Returns:
            Any: Selected data
        """
        self._control()
        self._query_control()
        query, data = build_select(
            self._table, where, what, 1, 0, order
        )  # type: ignore
        with self._sql:
            cursor = self._exec(query, data)
            returned = cursor.fetchone()
            if isinstance(what, ParsedFn):
                print(returned)
                return returned[what.parse_sql()[0]]
            if not returned:
                return Row()
            if isinstance(what, tuple) and len(what) == 1:
                return returned[what]
            if isinstance(what, str) and what != "*":
                return returned[what]
            return returned

    def columns(self):
        """Table columns"""
        if self._columns is None:
            raise AttributeError("columns are undefined.")

        return tuple(self._columns)

    def add_column(self, column: Column | BuilderColumn):
        """Add column to table"""
        sql = self._sql
        column = column.to_column() if isinstance(column, BuilderColumn) else column
        if column.primary or column.unique:
            raise OperationalError(
                "New column cannot have primary or unique constraint"
            )
        if column.nullable is False and column.default is None:
            raise OperationalError(
                "New column cannot be not null while default value is \
set to null"
            )
        if column.default is not None and column.foreign:
            raise OperationalError(
                "New column must accept null default value if foreign \
constraint is enabled."
            )
        query = f"alter table {self._table} add column {extract_single_column(column)}"
        if self._columns is not None:
            self._columns.append(column)
        sql.execute(query)

    def subquery(self, where: Condition, columns: OnlyColumn | str, limit: int = 0) -> SubQuery:
        """Push subquery to current .select() of other table"""
        return SubQuery(self, columns, where, limit)

    def rename_column(self, old_column: str, new_column: str):
        """Rename existing column to new column"""
        check_iter((old_column, new_column))
        query = f"alter table {self._table} rename column {old_column} to {new_column}"
        self._sql.execute(query)

    def commit(self):
        """Commit changes"""
        self._sql.commit()
        self._dirty = False

    def rollback(self):
        """Rollback"""
        self._sql.rollback()
        self._dirty = False

    def _begin_transaction(self):
        """Start a transaction or savepoint depending on depth."""
        stack = list(_tx_stack.get())  # copy since ContextVar values are immutable
        depth = len(stack)

        if depth == 0:
            self._sql.execute("BEGIN TRANSACTION")
        else:
            savepoint_name = f"sp_{depth}"
            self._sql.execute(f"SAVEPOINT {savepoint_name}")

        stack.append(True)
        _tx_stack.set(stack)

    def _commit_transaction(self):
        """Commit or release savepoint depending on depth."""
        stack = list(_tx_stack.get())
        depth = len(stack)

        if depth == 1:
            self._sql.commit()
        elif depth > 1:
            savepoint_name = f"sp_{depth-1}"
            self._sql.execute(f"RELEASE SAVEPOINT {savepoint_name}")
        stack.pop()
        _tx_stack.set(stack)

    def _rollback_transaction(self):
        """Rollback or rollback to savepoint depending on depth."""
        stack = list(_tx_stack.get())
        depth = len(stack)

        if depth == 1:
            self._sql.rollback()
        elif depth > 1:
            savepoint_name = f"sp_{depth-1}"
            self._sql.execute(f"ROLLBACK TO SAVEPOINT {savepoint_name}"
            )

        stack.pop()
        _tx_stack.set(stack)

    def count(self):
        """Count how much objects/rows stored in this table"""
        # ? Might as well uses __len__? But it's quite expensive.
        return self.select(what=count("*"))

    def __repr__(self) -> str:
        return f"<{type(self).__name__}({self._table}) -> {self._db!r}>"

__all__ = ["Table"]



"""Database"""

from . import models
from .models import BaseModel, model, Foreign, Primary, Unique, hook, validate
from .database import Database
from .utils import Row, SKIP
from .column import Column, text, integer, blob, real
from .signature import op
from .operators import this
from .table import Table
from .index import Index

def test_installed():
    """Is the module installed?"""
    return True


__version__ = "0.8.0"
__all__ = [
    "Database",
    "Table",
    "this",
    "op",
    "Column",
    "SKIP",
    "Row",
    "text",
    "integer",
    "real",
    "blob",
    "BaseModel",
    "models",
    "Foreign",
    "Primary",
    "Unique",
    "model",
    "hook",
    "validate",
    "Index"
]



"""SQLite Functions, this provide 'low level interface'."""

from hashlib import md5
from typing import NamedTuple, Any

from .utils import check_one  # type: ignore


def get_uid_from_args(fn_name, args):
    """Get unique id suffix from args"""
    hasher = md5(usedforsecurity=False)
    hasher.update(fn_name.encode())
    for arg in args:
        hasher.update(str(arg).encode())
    return int(hasher.hexdigest()[:8], 16)

# It was moved here, so users don't just import anything from query builder.


def _function_extract(
    parsed: "ParsedFn", depth: int = 0
) -> tuple[str, dict[str, Any]]:
    """Extract function into SQL function syntax."""
    check_one(parsed.name)
    if len(parsed.values) == 1:
        if parsed.values[0] in (None, ..., "*"):
            return f"{parsed.name}(*)", {}
        # need to find a check if input is column name or some values.
        return f"{parsed.name}({check_one(parsed.values[0])})", {}
    string = parsed.name + "("
    data = {}
    # ? we don't know how many same function calls at the same time, though we can use count param.
    suffix = get_uid_from_args(parsed.name, parsed.values)
    for i, _ in enumerate(parsed.values):
        key = f"fncall_{parsed.name}_{depth}_{i}__{suffix}"
        if isinstance(i, ParsedFn):
            sql, databin = _function_extract(i)
            data[key] = sql
            data.update(databin)
            continue
        data[key] = i
        string += f":{key}, "
    string = string[:-2] + ")"
    return string, data


class ParsedFn(NamedTuple):
    """Next step to query_builder"""

    name: str
    values: tuple[Any, ...]

    def parse_sql(self):
        """Parse the ParsedFn into SQL"""
        return _function_extract(self)


class Function:  # pylint: disable=too-few-public-methods
    """Provide Function API calls"""

    def __init__(self, name: str):
        self._name = check_one(name)

    def __call__(self, *args):
        return ParsedFn(self._name, args)

    @property
    def name(self):
        """Name of the function"""
        return self._name


count = Function("COUNT")



"""Errors"""

class Rejection(RuntimeError):
    """Cannot push into worker db"""

class DependencyError(ImportError):
    """Specific dependency is missing"""


class SecurityError(Exception):
    """SQLInjection ahead."""


class UnexpectedResultError(Exception):
    """Result of something is unexpected."""


class TableRemovedError(Exception):
    """Table is removed"""


class DatabaseExistsError(UnexpectedResultError):
    """Database that would be created is already exists."""


class DatabaseMissingError(UnexpectedResultError):
    """Database that would be accessed is missing."""


class ObjectRemovedError(BaseException):
    """Object is deleted from memory and cannot be obtained"""

class CuteDemonLordException(Exception):
    """A demon lord (cute one, somehow a dude) aborts an SQL statement"""

class FeatureGatekeep(ValueError):
    """A feature is gatekeeped"""

class VersionError(FeatureGatekeep):
    """Current version cannot use this feature"""

class ImplementationWarning(FutureWarning):
    """The implementation of specific feature is not supported"""



# pylint: disable=all
from luminadb.query_builder.table_creation import extract_table

SQL = "CREATE TABLE tbl (row1 text not null, row2 text not null, foreign key (row2) references tbl (row1) on delete cascade on update cascade)"


def test_query_builder_table_01():
    data = extract_table(SQL)
    for col in data:
        assert col.nullable is False
        if col.name == "row2":
            assert col.raw_source == "tbl/row1"




"""Utility"""

from typing import Optional, Any

from ..typings import Orders, Data
from ..functions import ParsedFn, _function_extract
from ..utils import check_one, check_iter, SKIP
from ..locals import _SQLITETYPES

from .typings import Condition, CacheOrders, CacheData


DEFAULT_MAPPINGS = {value: value for value in _SQLITETYPES}
SQL_ACTIONS = {"null": "set null"}
MAX_SUBQUERY_STACK_LIMIT = 10

NAMING_FORMAT = "{key}{suffix}__{call_id}_{depth}_{condition_id}"


def set_subquery_stack_limit(value: int):
    """Set subquery stack limit"""
    global MAX_SUBQUERY_STACK_LIMIT  # pylint: disable=global-statement
    if MAX_SUBQUERY_STACK_LIMIT <= 0:
        raise ValueError("Cannot set limit below 1")
    MAX_SUBQUERY_STACK_LIMIT = value


def select_onlyparam_parse(data: str | ParsedFn):
    """Select() parse `what` parameter"""
    if isinstance(data, str):
        return data
    x = _function_extract(data)
    if isinstance(x, tuple):
        return x[0]
    return x


def setup_hashable(
    condition: Condition, order: Optional[Orders] = None, data: Data | None = None
):
    """Setup hashable conditions, orders, and data"""
    cond = None
    order_ = None
    data_ = ()
    if isinstance(condition, dict):
        cond = tuple(condition.items())
    if isinstance(condition, list):
        cond = tuple(condition)

    if isinstance(order, tuple):
        order_ = order

    if data:
        data_ = tuple(data.keys())
    return cond, order_, data_


def setup_limit_patch(table_name: str, condition: str, limit):
    """Setup limit for patch/update"""
    check_one(table_name)
    if not isinstance(limit, int):
        limit = 1
    return f"where rowid in (select rowid from {table_name}\
{' '+condition if condition else ''} limit {limit})"


def parse_orders(order: CacheOrders):
    """Parse `orders` param"""
    if isinstance(order, tuple) and not isinstance(order[0], tuple):
        ord_, order_by = order
        # print('here')
        check_iter(order, ("asc", "desc"))  # type: ignore
        return f"{ord_} {order_by}"
    if isinstance(order, tuple) and isinstance(order[0], tuple):

        return ", ".join(
            f"{ord_} {order_by}"
            for ord_, order_by in order
            if check_iter((ord_, order_by), ("asc", "desc"))
        )
    raise TypeError("What?", type(order))


def remove_null(condition: dict[str, Any]) -> dict[str, Any]:
    """Remove SKIP from current Data"""
    new = condition.copy()
    for key, value in condition.items():
        if value is SKIP:
            del new[key]
    if not new:
        raise ValueError(
            "After removing SKIP sentinel value, new data that would be inserted"
            "/updated returns empty dictionary."
        )
    return new


def filter_extraction(string: str, shlexed: list[str]):
    """
    A function step of table extraction. Used to replace quoted and parens with parameter.

    Args:
        string (str): The input string containing the table creation SQL.
        shlexed (list[str]): The tokenized list of the input string.

    Returns:
        tuple[dict[str, str], dict[str, str], str]: A tuple containing:
            - A dictionary mapping placeholders to quoted strings.
            - A dictionary mapping placeholders to parenthesized strings.
            - The modified string with placeholders.
    """

    quoted_wrap = {}
    paren_wrap = {}
    new_string = string
    for index, shlex_string in enumerate(shlexed):
        # ? this attempt to replace all quoted string to a format-able stuff.
        if shlex_string.startswith("'") or shlex_string.startswith('"'):
            quoted_wrap[f":index{index}"] = shlex_string
            new_string = new_string.replace(shlex_string, f":index{index}")

    while (index := new_string.find("(")) != -1:
        # ? this attempt to replace all brackets with format-able stuff
        last = new_string.find(")") + 1
        selected = new_string[index:last]
        new_string = new_string.replace(selected, f":wrap{len(paren_wrap)}")
        paren_wrap[f":wrap{len(paren_wrap)}"] = selected
    return quoted_wrap, paren_wrap, new_string


def format_paramable(data: dict[str, Any] | tuple[str, ...]):
    """Format a data to parameterized data."""
    that: dict[str, str] = {}
    if isinstance(data, dict):
        for key in data:
            check_one(key)
            that[key] = f":{key}"
    else:
        for key in data:
            check_one(key)
            that[key] = f":{key}"
    return that


def combine_keyvals(keydict: dict[str, Any], valuedict: dict[str, Any] | CacheData):
    """Combine key dictionary with value dictionary. The first dictionary will only
    ignore the values while value dict ignore the keys.
    Mapping[key, _] -> keydict
    Mapping[_, value] -> valuedict"""
    if len(keydict) != len(valuedict):
        raise IndexError("One dictionary is larger. It must be equal.")
    new: dict[str, Any] = {}
    if isinstance(valuedict, dict):
        for key0, key1 in zip(keydict, valuedict):
            new[key0] = valuedict[key1]
    else:
        for key0, val1 in zip(keydict, valuedict):
            new[key0] = val1
    return new



"""Typings"""

from typing import TypeAlias, Any, Literal

from ..subquery import SubQuery
from ..functions import ParsedFn
from ..signature import Signature

ConditionDict: TypeAlias = dict[str, Signature | ParsedFn | SubQuery | Any]
ConditionList: TypeAlias = list[tuple[str, Signature | SubQuery | ParsedFn]]
Condition: TypeAlias = ConditionDict | ConditionList | None
CacheCond: TypeAlias = tuple[tuple[str, Signature | SubQuery | ParsedFn], ...] | None
CacheOrders: TypeAlias = tuple[str, Literal["asc", "desc"]] | ParsedFn | None
CacheData: TypeAlias = tuple[str, ...]
OnlyColumn: TypeAlias = tuple[str, ...] | str | ParsedFn



"""Model API tests"""

from uuid import UUID, uuid4
from pytest import raises
from luminadb import Database, model, Primary, Unique, BaseModel
from luminadb.models import hook, validate
from luminadb.models.mixin import ScopeMixin, ChunkableMixin
from luminadb.models.errors import ValidationError, NoDataReturnedError

from ..setup import setup_model_api

def auto_id():
    return str(uuid4())

def test_model_api():
    # pylint: disable=protected-access
    """Test 1100 Model API"""
    db = Database(":memory:")

    @model(db)
    class Users(BaseModel):  # type: ignore
        """Users"""

        __schema__ = (Primary("id"), Unique("username"))
        id: str
        username: str
        display_name: str
        is_active: bool = True

    assert db.table("users")._table == Users._tbl._table
    admin = Users.create(
        id=str(UUID(int=0)), username="admin", display_name="System Administrator"
    )
    assert admin.username
    fetched = Users.where(username="admin").fetch_one()
    assert fetched == admin


def test_model_relationship():
    """Test 1101 Model API Relationship"""

    db = Database(":memory:")

    Users, Posts = setup_model_api(db)

    admin = Users.create(id="0", username="Admin")
    post0 = Posts.create(
        id="0",
        title="Hello, World!",
        content="Lorem Ipsum Dolor sit Amet",
        user_id=admin.id,
    )
    user0 = Posts.belongs_to(post0, Users)
    assert admin == user0, "belongs_to() should return the correct user"
    assert post0 in admin.has_many(
        Posts
    ), "has_many() should return related posts to user"


def test_model_mixin():
    """Test 1102 Model API Mixins"""

    db = Database(":memory:")

    @model(db)
    class Posts(BaseModel, ScopeMixin, ChunkableMixin):
        """Posts"""

        __schema__ = (Primary("id"),)

        id: str
        title: str
        content: str
        is_active: bool

    Posts.bulk_create(
        [
            {"id": str(uuid4()), "title": "a", "content": "a", "is_active": True}
            for _ in range(10)
        ]
    )

    for posts in Posts.chunk(5, title="a"):
        assert all(map(lambda post: post.title, posts))

    assert Posts.active()

def test_model_hooks_usability():
    """Test Model API if hooks can be used"""

    db = Database(":memory:")
    called = {"before_create": False, "after_create": False}

    @model(db)
    class Users(BaseModel):
        """Base User class"""

        id: str
        username: str
        is_active: bool

        @hook
        def before_create(_):
            """Before create"""
            called["before_create"] = True

        @hook("after_create")
        def hook0(_):
            """After create"""
            called['after_create'] = True

    assert hasattr(Users, "before_create")

    Users.create(id="2", username="admin", is_active=True)

    assert called['before_create'] is True, "Hooks is not called"
    assert called['after_create'] is True, "Hooks is not called"

def test_model_hooks_and_validator():
    """Test Model API Validators"""

    db = Database(":memory:")

    @model(db)
    class Users(BaseModel):
        """Base User class"""

        id: str
        username: str
        is_active: bool

        @validate
        def validate_is_active(instance: "Users"):  # type: ignore
            """Is active must be a boolean"""
            return isinstance(instance.is_active, bool)

        @validate("username", "Username must be admin")
        def is_admin_vld(instance: "Users"):
            """Username must be admin"""
            return instance.username == 'admin'

    assert hasattr(Users, "validate_is_active")

    with raises(ValidationError):
        Users.create(id="0", username="admin", is_active=7773)
    Users.create(id="0", username="admin", is_active=True)

def test_model_auto_id():
    """Test 1104 Model API Auto ID"""

    db = Database(":memory:")

    @model(db)
    class Users(BaseModel):
        """Base User class"""

        __schema__ = (Primary("id"),)
        __auto_id__ = auto_id
        id: str
        username: str
        is_active: bool

        # This also works!
        # @staticmethod
        # def __auto_id__():
        #     return str(uuid4())

    assert Users.create(username="admin", is_active=False).id

def test_model_hidden():
    """Test 1105 Model API __hidden__"""

    db = Database(":memory:")

    @model(db)
    class Users(BaseModel):
        """Base User class"""

        __schema__ = (Primary("id"),)
        __auto_id__ = auto_id
        __hidden__ = ("password",)
        id: str
        username: str
        password: str

    admin = Users.create(username="admin", password="admin123")
    admin_dict = admin.to_dict()
    assert "password" not in admin_dict
    assert admin.to_safe_instance().password is None

def test_model_fail():
    """Test 1106 Model API fail-able methods"""

    db = Database(":memory:")

    @model(db)
    class Users(BaseModel):
        """Base User class"""

        __schema__ = (Primary("id"),)
        __auto_id__ = auto_id
        __hidden__ = ("password",)
        id: str
        username: str
        password: str

    assert Users.first() is None, "Users class has unexpected data"
    with raises(NoDataReturnedError):
        Users.find_or_fail(1)

def test_model_api_runtime_typechecking():
    """Test Model API runtime type checking"""

    db = Database(":memory:")

    @model(db, True)
    class Users(BaseModel):
        """Base User class"""

        __schema__ = (Primary("id"),)
        __auto_id__ = auto_id
        __hidden__ = ("password",)
        id: str
        username: str
        password: str

    with raises(ValidationError):
        Users.create(id=0, username=0, password=0)

    assert Users.first() is None, "How?"

def test_model_api_query_builder_update():
    """Test Model API Querybuilder update()"""

    db = Database(":memory:")

    Users, _ = setup_model_api(db)
    admin = Users.create(id=auto_id(), username="admin", is_active=True)
    assert Users.where(id=admin.id).throw().patch(username="system") == 1, "Changes should be 1"

def test_model_api_query_builder_delete():
    """Test Model API QueryBuilder delete()"""

    db = Database(":memory:")

    Users, _ = setup_model_api(db)
    admin = Users.create(id=auto_id(), username='admin', is_active=True)

    assert Users.where(id=admin.id).throw().delete() == 1, "Changes should be 1"

def test_model_api_custom_id():
    """Test Model API Custom Primary ID"""

    db = Database(":memory:")
    db.foreign_pragma('ON')

    @model(db)
    class Users(BaseModel):
        __schema__ = (Primary('uid'),)

        uid: str
        username: str

    assert Users.create(uid="1", username='admin')



"""Test Table API select"""

from luminadb import Database, text, integer
from luminadb.operators import op, eq

from ..setup import (
    GROUP_BASE,
    USER_BASE,
    save_report,
    setup_database,
    setup_orderable,
    USER_NAMEBASE,
    GROUP_BASE_CRUNCHED,
    setup_database_builder,
    setup_database_fns,
    temp_dir
)

database = Database(temp_dir / "test-selection.db")  # type: ignore
setup_database_builder(database)
setup_database_fns(database)
users = database.table("users")
groups = database.table("groups")


def test_select():
    """test 0000 select"""
    assert groups.select_one({"id": op == 0}) == GROUP_BASE[0]
    assert groups.select() == GROUP_BASE
    assert users.select_one({"id": op == 0}) == USER_BASE[0]
    assert users.select() == USER_BASE
    assert save_report("00_test", database, groups, users)


def test_00_01_select():
    """Test 0001 select"""
    assert groups.select_one([eq("id", 0)]) == GROUP_BASE[0]
    assert groups.select() == GROUP_BASE
    assert users.select_one([eq("id", 0)]) == USER_BASE[0]
    assert users.select() == USER_BASE
    assert save_report("00_test", database, groups, users)


def test_00_02_select():
    """Test 0002 select"""
    assert groups.select_one({"id": 0}) == GROUP_BASE[0]
    assert groups.select() == GROUP_BASE
    assert users.select_one({"id": 0}) == USER_BASE[0]
    assert users.select() == USER_BASE
    assert save_report("00_test", database, groups, users)

    assert users.select_one({"id": 0}, ("username", "role")) == USER_NAMEBASE[0]
    assert save_report("00_test", database, groups, users)


def test_select_crunch():
    """Test 0004 select with crunch"""
    db = Database(":memory:")
    setup_database(db)
    gr = db.table("groups")
    assert gr.select(flatten=True) == GROUP_BASE_CRUNCHED


def test_select_one_only_one_item():
    """Test 0005"""
    db = Database(":memory:")
    setup_database(db)
    gr = db.table("groups")
    assert gr.select_one({"id": 0}, "name") == "root"


def test_select_order_by():
    """Test 0006 Select order by"""
    db = Database(":memory:")
    setup_orderable(db)
    items = db.table("items")
    first_high = items.select(what="quantity", limit=3, order=("quantity", "asc"))
    first_low = items.select(what="quantity", limit=3, order=("quantity", "desc"))
    assert first_high == [0, 1, 2]
    assert first_low == [99, 98, 97]

def test_select_subquery():
    """Test 0007 Select with subqueries"""
    db = Database(":memory:")
    notes = db.create_table("notes", [text("id").primary(), text("content")])
    notes.insert({"id": "0", "content": "abc"})
    data = notes.select_one({"id": notes.subquery({"id": "0"}, "id", limit=1)})
    assert data is not None

def test_paginate_select():
    """Test 0500 Pagination select"""
    data = []
    for a, b in zip(range(0, 100), range(1000, 1100)):
        data.append({"x": a, "y": b})

    db = Database(":memory:")
    nums = db.create_table("nums", [integer("x"), integer("y")])

    nums.insert_many(data)

    for i in nums.paginate_select():
        assert i



"""Typing Extensions"""

# pylint: disable=unnecessary-ellipsis
from typing import Any, Literal, TypedDict, TypeAlias, TypeVar
from .utils import Queries, SquashedQueries, Query


T = TypeVar("T")
Order: TypeAlias = tuple[str, Literal['asc'] | Literal['desc']]
Orders: TypeAlias = tuple[Order, ...] | Order
Data: TypeAlias = dict[str, Any]
OnlyColumn: TypeAlias = tuple[str, ...] | Literal["*"]
JustAColumn: TypeAlias = str | tuple[str] # pylint: disable=invalid-name
tuple_list: TypeAlias = list[T] | tuple[T, ...] # pylint: disable=invalid-name
null = object()


class MasterQuery(TypedDict):
    """Master Query"""

    type: str
    name: str
    tbl_name: str
    rootpage: int
    sql: str

__all__ = (
    "Order",
    "Orders",
    "Data",
    'OnlyColumn',
    "JustAColumn",
    "tuple_list",
    "Queries",
    "SquashedQueries",
    "Query",
    "null"
)
